@@ -1,7 +1,7 @@
 /*****************************************************************************
  * dvb.c: linux-dvb input for DVBlast
  *****************************************************************************
- * Copyright (C) 2008-2010 VideoLAN
+ * Copyright (C) 2008-2010, 2015 VideoLAN
  *
  * Authors: Christophe Massiot <massiot@via.ecp.fr>
  *
@@ -19,6 +19,9 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
  *****************************************************************************/
+#include "config.h"
+
+#ifdef HAVE_DVB_SUPPORT
 
 #include <stdlib.h>
 #include <stdint.h>
@@ -44,6 +47,15 @@
 #include <linux/dvb/frontend.h>
 #include <linux/dvb/ca.h>
 
+#include <ev.h>
+
+#if DVBAPI_VERSION < 508
+  #define DTV_STREAM_ID        42
+  #define FE_CAN_MULTISTREAM   0x4000000
+#endif
+
+#define MAX_DELIVERY_SYSTEMS 20
+
 #include "dvblast.h"
 #include "en50221.h"
 #include "comm.h"
@@ -54,22 +66,24 @@
  * Local declarations
  *****************************************************************************/
 #define DVR_READ_TIMEOUT 30000000 /* 30 s */
-#define CA_POLL_PERIOD 100000 /* 100 ms */
 #define MAX_READ_ONCE 50
 #define DVR_BUFFER_SIZE 40*188*1024 /* bytes */
 
+int i_dvr_buffer_size = DVR_BUFFER_SIZE;
+
 static int i_frontend, i_dvr;
+static struct ev_io frontend_watcher, dvr_watcher;
+static struct ev_timer lock_watcher, mute_watcher, print_watcher;
 static fe_status_t i_last_status;
-static mtime_t i_frontend_timeout;
-static mtime_t i_last_packet = 0;
-static mtime_t i_ca_next_event = 0;
 static block_t *p_freelist = NULL;
 
 /*****************************************************************************
  * Local prototypes
  *****************************************************************************/
-static block_t *DVRRead( void );
-static void FrontendPoll( void );
+static void DVRRead(struct ev_loop *loop, struct ev_io *w, int revents);
+static void DVRMuteCb(struct ev_loop *loop, struct ev_timer *w, int revents);
+static void FrontendRead(struct ev_loop *loop, struct ev_io *w, int revents);
+static void FrontendLockCb(struct ev_loop *loop, struct ev_timer *w, int revents);
 static void FrontendSet( bool b_reset );
 
 /*****************************************************************************
@@ -79,19 +93,24 @@ void dvb_Open( void )
 {
     char psz_tmp[128];
 
-    msg_Dbg( NULL, "using linux-dvb API version %d", DVB_API_VERSION );
+    msg_Dbg( NULL, "compiled with DVB API version %d.%d", DVB_API_VERSION, DVB_API_VERSION_MINOR );
 
-    i_wallclock = mdate();
-
-    sprintf( psz_tmp, "/dev/dvb/adapter%d/frontend%d", i_adapter, i_fenum );
-    if( (i_frontend = open(psz_tmp, O_RDWR | O_NONBLOCK)) < 0 )
+    if ( i_frequency )
     {
-        msg_Err( NULL, "opening device %s failed (%s)", psz_tmp,
-                 strerror(errno) );
-        exit(1);
-    }
+        sprintf( psz_tmp, "/dev/dvb/adapter%d/frontend%d", i_adapter, i_fenum );
+        if( (i_frontend = open(psz_tmp, O_RDWR | O_NONBLOCK)) < 0 )
+        {
+            msg_Err( NULL, "opening device %s failed (%s)", psz_tmp,
+                     strerror(errno) );
+            exit(1);
+        }
 
-    FrontendSet(true);
+        FrontendSet(true);
+    }
+    else
+    {
+        i_frontend = -1;
+    }
 
     sprintf( psz_tmp, "/dev/dvb/adapter%d/dvr%d", i_adapter, i_fenum );
 
@@ -102,14 +121,29 @@ void dvb_Open( void )
         exit(1);
     }
 
-    if ( ioctl( i_dvr, DMX_SET_BUFFER_SIZE, DVR_BUFFER_SIZE ) < 0 )
+    if ( ioctl( i_dvr, DMX_SET_BUFFER_SIZE, i_dvr_buffer_size ) < 0 )
     {
         msg_Warn( NULL, "couldn't set %s buffer size (%s)", psz_tmp,
                  strerror(errno) );
     }
 
+    ev_io_init(&dvr_watcher, DVRRead, i_dvr, EV_READ);
+    ev_io_start(event_loop, &dvr_watcher);
+
+    if ( i_frontend != -1 )
+    {
+        ev_io_init(&frontend_watcher, FrontendRead, i_frontend, EV_READ);
+        ev_io_start(event_loop, &frontend_watcher);
+    }
+
+    ev_timer_init(&lock_watcher, FrontendLockCb,
+                  i_frontend_timeout_duration / 1000000.,
+                  i_frontend_timeout_duration / 1000000.);
+    ev_timer_init(&mute_watcher, DVRMuteCb,
+                  DVR_READ_TIMEOUT / 1000000.,
+                  DVR_READ_TIMEOUT / 1000000.);
+
     en50221_Init();
-    i_ca_next_event = mdate() + CA_POLL_PERIOD;
 }
 
 /*****************************************************************************
@@ -117,108 +151,14 @@ void dvb_Open( void )
  *****************************************************************************/
 void dvb_Reset( void )
 {
-    FrontendSet(true);
+    if ( i_frequency )
+        FrontendSet(true);
 }
 
 /*****************************************************************************
- * dvb_Read
+ * DVR events
  *****************************************************************************/
-block_t *dvb_Read( mtime_t i_poll_timeout )
-{
-    struct pollfd ufds[4];
-    int i_ret, i_nb_fd = 2;
-    block_t *p_blocks = NULL;
-
-    memset( ufds, 0, sizeof(ufds) );
-    ufds[0].fd = i_dvr;
-    ufds[0].events = POLLIN;
-    ufds[1].fd = i_frontend;
-    ufds[1].events = POLLERR | POLLPRI;
-    if ( i_comm_fd != -1 )
-    {
-        ufds[i_nb_fd].fd = i_comm_fd;
-        ufds[i_nb_fd].events = POLLIN;
-        i_nb_fd++;
-    }
-    if ( i_ca_handle && i_ca_type == CA_CI_LINK )
-    {
-        ufds[i_nb_fd].fd = i_ca_handle;
-        ufds[i_nb_fd].events = POLLIN;
-        i_nb_fd++;
-    }
-
-    i_ret = poll( ufds, i_nb_fd, (i_poll_timeout + 999) / 1000 );
-
-    i_wallclock = mdate();
-
-    if ( i_ret < 0 )
-    {
-        if( errno != EINTR )
-            msg_Err( NULL, "poll error: %s", strerror(errno) );
-        return NULL;
-    }
-
-    if ( ufds[1].revents )
-        FrontendPoll();
-
-    if ( ufds[0].revents )
-    {
-        p_blocks = DVRRead();
-        i_wallclock = mdate();
-    }
-
-    if ( p_blocks != NULL )
-        i_last_packet = i_wallclock;
-    else if ( !i_frontend_timeout
-                && i_wallclock > i_last_packet + DVR_READ_TIMEOUT )
-    {
-        msg_Warn( NULL, "no DVR output, resetting" );
-        FrontendSet(false);
-        en50221_Reset();
-    }
-
-    if ( i_ca_handle && i_ca_type == CA_CI_LINK )
-    {
-        if ( ufds[i_nb_fd - 1].revents )
-        {
-            en50221_Read();
-            i_ca_next_event = i_wallclock + CA_POLL_PERIOD;
-        }
-        else if ( i_wallclock > i_ca_next_event )
-        {
-            en50221_Poll();
-            i_ca_next_event = i_wallclock + CA_POLL_PERIOD;
-        }
-    }
-
-    if ( i_frontend_timeout && i_wallclock > i_frontend_timeout )
-    {
-        if ( i_quit_timeout_duration )
-        {
-            msg_Err( NULL, "no lock" );
-            switch (i_print_type) {
-            case PRINT_XML:
-                printf("</TS>\n");
-                break;
-            default:
-                break;
-            }
-            exit(EXIT_STATUS_FRONTEND_TIMEOUT);
-        }
-        msg_Warn( NULL, "no lock, tuning again" );
-        FrontendSet(false);
-    }
-
-    if ( i_comm_fd != -1 && ufds[2].revents )
-        comm_Read();
-
-    return p_blocks;
-}
-
-/*****************************************************************************
- * DVRRead
- *****************************************************************************/
-static block_t *DVRRead( void )
+static void DVRRead(struct ev_loop *loop, struct ev_io *w, int revents)
 {
     int i, i_len;
     block_t *p_ts = p_freelist, **pp_current = &p_ts;
@@ -240,6 +180,9 @@ static block_t *DVRRead( void )
     }
     i_len /= TS_SIZE;
 
+    if ( i_len )
+        ev_timer_again(loop, &mute_watcher);
+
     pp_current = &p_ts;
     while ( i_len && *pp_current )
     {
@@ -250,7 +193,27 @@ static block_t *DVRRead( void )
     p_freelist = *pp_current;
     *pp_current = NULL;
 
-    return p_ts;
+    demux_Run( p_ts );
+}
+
+static void DVRMuteCb(struct ev_loop *loop, struct ev_timer *w, int revents)
+{
+    msg_Warn( NULL, "no DVR output, resetting" );
+    ev_timer_stop(loop, w);
+
+    switch (i_print_type) {
+    case PRINT_XML:
+        fprintf(print_fh, "<EVENT type=\"reset\" cause=\"dvr\" />\n");
+        break;
+    case PRINT_TEXT:
+        fprintf(print_fh, "reset cause: dvr\n");
+        break;
+    default:
+        break;
+    }
+    if ( i_frequency )
+        FrontendSet(false);
+    en50221_Reset();
 }
 
 
@@ -313,9 +276,39 @@ void dvb_UnsetFilter( int i_fd, uint16_t i_pid )
  */
 
 /*****************************************************************************
- * FrontendPoll : Poll for frontend events
+ * Print info
  *****************************************************************************/
-static void FrontendPoll( void )
+static void PrintCb( struct ev_loop *loop, struct ev_timer *w, int revents )
+{
+    uint32_t i_ber = 0;
+    uint16_t i_strength = 0, i_snr = 0;
+    uint32_t i_uncorrected = 0;
+
+    ioctl(i_frontend, FE_READ_BER, &i_ber);
+    ioctl(i_frontend, FE_READ_SIGNAL_STRENGTH, &i_strength);
+    ioctl(i_frontend, FE_READ_SNR, &i_snr);
+    ioctl(i_frontend, FE_READ_UNCORRECTED_BLOCKS, &i_uncorrected);
+
+    switch (i_print_type)
+    {
+        case PRINT_XML:
+            fprintf(print_fh,
+                    "<STATUS type=\"frontend\" ber=\"%"PRIu32"\" strength=\"%"PRIu16"\" snr=\"%"PRIu16"\" uncorrected=\"%"PRIu32"\" />\n",
+                    i_ber, i_strength, i_snr, i_uncorrected);
+            break;
+        case PRINT_TEXT:
+            fprintf(print_fh, "frontend ber: %"PRIu32" strength: %"PRIu16" snr: %"PRIu16" uncorrected: %"PRIu32"\n",
+                    i_ber, i_strength, i_snr, i_uncorrected);
+            break;
+        default:
+            break;
+    }
+}
+
+/*****************************************************************************
+ * Frontend events
+ *****************************************************************************/
+static void FrontendRead(struct ev_loop *loop, struct ev_io *w, int revents)
 {
     struct dvb_frontend_event event;
     fe_status_t i_status, i_diff;
@@ -371,16 +364,17 @@ static void FrontendPoll( void )
                 msg_Info( NULL, "frontend has acquired lock" );
                 switch (i_print_type) {
                 case PRINT_XML:
-                    printf("<STATUS type=\"lock\" status=\"1\" />\n");
+                    fprintf(print_fh, "<STATUS type=\"lock\" status=\"1\" />\n");
+                    break;
+                case PRINT_TEXT:
+                    fprintf(print_fh, "lock status: 1\n");
                     break;
                 default:
-                    printf("frontend has acquired lock\n" );
+                    break;
                 }
-                i_frontend_timeout = 0;
-                i_last_packet = i_wallclock;
 
-                if ( i_quit_timeout_duration && !i_quit_timeout )
-                    i_quit_timeout = i_wallclock + i_quit_timeout_duration;
+                ev_timer_stop(loop, &lock_watcher);
+                ev_timer_again(loop, &mute_watcher);
 
                 /* Read some statistics */
                 if( ioctl( i_frontend, FE_READ_BER, &i_value ) >= 0 )
@@ -389,31 +383,77 @@ static void FrontendPoll( void )
                     msg_Dbg( NULL, "- Signal strength: %d", i_value );
                 if( ioctl( i_frontend, FE_READ_SNR, &i_value ) >= 0 )
                     msg_Dbg( NULL, "- SNR: %d", i_value );
+
+                if (i_print_period)
+                {
+                    ev_timer_init( &print_watcher, PrintCb,
+                                   i_print_period / 1000000.,
+                                   i_print_period / 1000000. );
+                    ev_timer_start( event_loop, &print_watcher );
+                }
             }
             else
             {
                 msg_Dbg( NULL, "frontend has lost lock" );
                 switch (i_print_type) {
                 case PRINT_XML:
-                    printf("<STATUS type=\"lock\" status=\"0\"/>\n");
+                    fprintf(print_fh, "<STATUS type=\"lock\" status=\"0\"/>\n");
+                    break;
+                case PRINT_TEXT:
+                    fprintf(print_fh, "lock status: 0\n");
                     break;
                 default:
-                    printf("frontend has lost lock\n" );
+                    break;
                 }
-                i_frontend_timeout = i_wallclock + i_frontend_timeout_duration;
+
+                if (i_frontend_timeout_duration)
+                {
+                    ev_timer_stop(event_loop, &lock_watcher);
+                    ev_timer_again(loop, &mute_watcher);
+                }
+
+                if (i_print_period)
+                    ev_timer_stop(event_loop, &print_watcher);
             }
 
             IF_UP( FE_REINIT )
             {
                 /* The frontend was reinited. */
                 msg_Warn( NULL, "reiniting frontend");
-                FrontendSet(true);
+                if ( i_frequency )
+                    FrontendSet(true);
             }
         }
 #undef IF_UP
     }
 }
 
+static void FrontendLockCb(struct ev_loop *loop, struct ev_timer *w, int revents)
+{
+    if ( i_quit_timeout_duration )
+    {
+        msg_Err( NULL, "no lock" );
+        ev_break(loop, EVBREAK_ALL);
+        return;
+    }
+
+    msg_Warn( NULL, "no lock, tuning again" );
+    ev_timer_stop(loop, w);
+
+    switch (i_print_type) {
+    case PRINT_XML:
+        fprintf(print_fh, "<EVENT type=\"reset\" cause=\"nolock\" />\n");
+        break;
+    case PRINT_TEXT:
+        fprintf(print_fh, "reset cause: nolock\n");
+        break;
+    default:
+        break;
+    }
+    if ( i_frequency )
+        FrontendSet(false);
+}
+
 static int FrontendDoDiseqc(void)
 {
     fe_sec_voltage_t fe_voltage;
@@ -495,13 +535,43 @@ static int FrontendDoDiseqc(void)
          * specification is available from http://www.eutelsat.com/
          */
 
+        /* DiSEqC 1.1 */
+        struct dvb_diseqc_master_cmd uncmd =
+            { {0xe0, 0x10, 0x39, 0xf0, 0x00, 0x00}, 4};
+
+        /* DiSEqC 1.0 */
         struct dvb_diseqc_master_cmd cmd =
             { {0xe0, 0x10, 0x38, 0xf0, 0x00, 0x00}, 4};
+
         cmd.msg[3] = 0xf0 /* reset bits */
                           | ((i_satnum - 1) << 2)
                           | (fe_voltage == SEC_VOLTAGE_13 ? 0 : 2)
                           | (fe_tone == SEC_TONE_ON ? 1 : 0);
 
+        if ( i_uncommitted > 0 && i_uncommitted < 5 )
+        {
+           uncmd.msg[3] = 0xf0 /* reset bits */
+                             | ((i_uncommitted - 1) << 2)
+                             | (fe_voltage == SEC_VOLTAGE_13 ? 0 : 2)
+                             | (fe_tone == SEC_TONE_ON ? 1 : 0);
+           if( ioctl( i_frontend, FE_DISEQC_SEND_MASTER_CMD, &uncmd ) < 0 )
+           {
+               msg_Err( NULL, "ioctl FE_SEND_MASTER_CMD failed (%s)",
+                        strerror(errno) );
+               exit(1);
+           }
+           /* Repeat uncommitted command */
+           uncmd.msg[0] = 0xe1; /* framing: master, no reply, repeated TX */
+           if( ioctl( i_frontend, FE_DISEQC_SEND_MASTER_CMD, &uncmd ) < 0 )
+           {
+               msg_Err( NULL, "ioctl FE_SEND_MASTER_CMD failed (%s)",
+                        strerror(errno) );
+               exit(1);
+           }
+           /* Pause 125 ms between uncommitted & committed diseqc commands. */
+           msleep(125000);
+        }
+
         if( ioctl( i_frontend, FE_DISEQC_SEND_MASTER_CMD, &cmd ) < 0 )
         {
             msg_Err( NULL, "ioctl FE_SEND_MASTER_CMD failed (%s)",
@@ -511,6 +581,7 @@ static int FrontendDoDiseqc(void)
         msleep(100000); /* Should be 15 ms. */
 
         /* Do it again just to be sure. */
+        cmd.msg[0] = 0xe1; /* framing: master, no reply, repeated TX */
         if( ioctl( i_frontend, FE_DISEQC_SEND_MASTER_CMD, &cmd ) < 0 )
         {
             msg_Err( NULL, "ioctl FE_SEND_MASTER_CMD failed (%s)",
@@ -539,15 +610,17 @@ static int FrontendDoDiseqc(void)
 
     msleep(100000); /* ... */
 
-    msg_Dbg( NULL, "configuring LNB to v=%d p=%d satnum=%x",
-             i_voltage, b_tone, i_satnum );
+    msg_Dbg( NULL, "configuring LNB to v=%d p=%d satnum=%x uncommitted=%x",
+             i_voltage, b_tone, i_satnum, i_uncommitted );
     return bis_frequency;
 }
 
-#define DVBAPI_VERSION ((DVB_API_VERSION)*100+(DVB_API_VERSION_MINOR))
-
 #if DVB_API_VERSION >= 5
 
+#if DVBAPI_VERSION < 505
+#warning Your linux-dvb headers are old, you should consider upgrading your kernel and/or compiling against different kernel headers
+#endif
+
 /*****************************************************************************
  * Helper functions for S2API
  *****************************************************************************/
@@ -693,78 +766,126 @@ static fe_hierarchy_t GetHierarchy(void)
 /*****************************************************************************
  * FrontendInfo : Print frontend info
  *****************************************************************************/
-static const char *GetFrontendTypeName( fe_type_t type )
+static void FrontendInfo( struct dvb_frontend_info *info, uint32_t version,
+                          fe_delivery_system_t *p_systems, int i_systems )
 {
-    switch(type)
-    {
-        case FE_QPSK: return "QPSK (DVB-S/S2)";
-        case FE_QAM:  return "QAM  (DVB-C)";
-        case FE_OFDM: return "OFDM (DVB-T)";
-        case FE_ATSC: return "ATSC";
-        default: return "unknown";
-    }
-}
-
-static void FrontendInfo( struct dvb_frontend_info info )
-{
-    msg_Dbg( NULL, "Frontend \"%s\" type \"%s\" supports:",
-             info.name, GetFrontendTypeName(info.type) );
+    msg_Dbg( NULL, "using DVB API version %d.%d", version / 256, version % 256 );
+    msg_Dbg( NULL, "Frontend \"%s\" supports:", info->name );
     msg_Dbg( NULL, " frequency min: %d, max: %d, stepsize: %d, tolerance: %d",
-             info.frequency_min, info.frequency_max,
-             info.frequency_stepsize, info.frequency_tolerance );
+             info->frequency_min, info->frequency_max,
+             info->frequency_stepsize, info->frequency_tolerance );
     msg_Dbg( NULL, " symbolrate min: %d, max: %d, tolerance: %d",
-             info.symbol_rate_min, info.symbol_rate_max, info.symbol_rate_tolerance);
+             info->symbol_rate_min, info->symbol_rate_max, info->symbol_rate_tolerance);
     msg_Dbg( NULL, " capabilities:" );
 
-#define FRONTEND_INFO(caps,val,msg)                                             \
+#define FRONTEND_INFO(caps,val,msg)                                         \
     if ( caps & val )                                                       \
         msg_Dbg( NULL, "  %s", msg );
 
-    FRONTEND_INFO( info.caps, FE_IS_STUPID, "FE_IS_STUPID" )
-    FRONTEND_INFO( info.caps, FE_CAN_INVERSION_AUTO, "INVERSION_AUTO" )
-    FRONTEND_INFO( info.caps, FE_CAN_FEC_1_2, "FEC_1_2" )
-    FRONTEND_INFO( info.caps, FE_CAN_FEC_2_3, "FEC_2_3" )
-    FRONTEND_INFO( info.caps, FE_CAN_FEC_3_4, "FEC_3_4" )
-    FRONTEND_INFO( info.caps, FE_CAN_FEC_4_5, "FEC_4_5" )
-    FRONTEND_INFO( info.caps, FE_CAN_FEC_5_6, "FEC_5_6" )
-    FRONTEND_INFO( info.caps, FE_CAN_FEC_6_7, "FEC_6_7" )
-    FRONTEND_INFO( info.caps, FE_CAN_FEC_7_8, "FEC_7_8" )
-    FRONTEND_INFO( info.caps, FE_CAN_FEC_8_9, "FEC_8_9" )
-    FRONTEND_INFO( info.caps, FE_CAN_FEC_AUTO,"FEC_AUTO")
-    FRONTEND_INFO( info.caps, FE_CAN_QPSK,   "QPSK" )
-    FRONTEND_INFO( info.caps, FE_CAN_QAM_16, "QAM_16" )
-    FRONTEND_INFO( info.caps, FE_CAN_QAM_32, "QAM_32" )
-    FRONTEND_INFO( info.caps, FE_CAN_QAM_64, "QAM_64" )
-    FRONTEND_INFO( info.caps, FE_CAN_QAM_128,"QAM_128")
-    FRONTEND_INFO( info.caps, FE_CAN_QAM_256,"QAM_256")
-    FRONTEND_INFO( info.caps, FE_CAN_QAM_AUTO,"QAM_AUTO" )
-    FRONTEND_INFO( info.caps, FE_CAN_TRANSMISSION_MODE_AUTO, "TRANSMISSION_MODE_AUTO" )
-    FRONTEND_INFO( info.caps, FE_CAN_BANDWIDTH_AUTO, "BANDWIDTH_AUTO" )
-    FRONTEND_INFO( info.caps, FE_CAN_GUARD_INTERVAL_AUTO, "GUARD_INTERVAL_AUTO" )
-    FRONTEND_INFO( info.caps, FE_CAN_HIERARCHY_AUTO, "HIERARCHY_AUTO" )
-    FRONTEND_INFO( info.caps, FE_CAN_8VSB, "8VSB" )
-    FRONTEND_INFO( info.caps, FE_CAN_16VSB,"16VSB" )
-    FRONTEND_INFO( info.caps, FE_HAS_EXTENDED_CAPS, "EXTENDED_CAPS" )
+    FRONTEND_INFO( info->caps, FE_IS_STUPID, "FE_IS_STUPID" )
+    FRONTEND_INFO( info->caps, FE_CAN_INVERSION_AUTO, "INVERSION_AUTO" )
+    FRONTEND_INFO( info->caps, FE_CAN_FEC_1_2, "FEC_1_2" )
+    FRONTEND_INFO( info->caps, FE_CAN_FEC_2_3, "FEC_2_3" )
+    FRONTEND_INFO( info->caps, FE_CAN_FEC_3_4, "FEC_3_4" )
+    FRONTEND_INFO( info->caps, FE_CAN_FEC_4_5, "FEC_4_5" )
+    FRONTEND_INFO( info->caps, FE_CAN_FEC_5_6, "FEC_5_6" )
+    FRONTEND_INFO( info->caps, FE_CAN_FEC_6_7, "FEC_6_7" )
+    FRONTEND_INFO( info->caps, FE_CAN_FEC_7_8, "FEC_7_8" )
+    FRONTEND_INFO( info->caps, FE_CAN_FEC_8_9, "FEC_8_9" )
+    FRONTEND_INFO( info->caps, FE_CAN_FEC_AUTO,"FEC_AUTO")
+    FRONTEND_INFO( info->caps, FE_CAN_QPSK,   "QPSK" )
+    FRONTEND_INFO( info->caps, FE_CAN_QAM_16, "QAM_16" )
+    FRONTEND_INFO( info->caps, FE_CAN_QAM_32, "QAM_32" )
+    FRONTEND_INFO( info->caps, FE_CAN_QAM_64, "QAM_64" )
+    FRONTEND_INFO( info->caps, FE_CAN_QAM_128,"QAM_128")
+    FRONTEND_INFO( info->caps, FE_CAN_QAM_256,"QAM_256")
+    FRONTEND_INFO( info->caps, FE_CAN_QAM_AUTO,"QAM_AUTO" )
+    FRONTEND_INFO( info->caps, FE_CAN_TRANSMISSION_MODE_AUTO, "TRANSMISSION_MODE_AUTO" )
+    FRONTEND_INFO( info->caps, FE_CAN_BANDWIDTH_AUTO, "BANDWIDTH_AUTO" )
+    FRONTEND_INFO( info->caps, FE_CAN_GUARD_INTERVAL_AUTO, "GUARD_INTERVAL_AUTO" )
+    FRONTEND_INFO( info->caps, FE_CAN_HIERARCHY_AUTO, "HIERARCHY_AUTO" )
+    FRONTEND_INFO( info->caps, FE_CAN_8VSB, "8VSB" )
+    FRONTEND_INFO( info->caps, FE_CAN_16VSB,"16VSB" )
+    FRONTEND_INFO( info->caps, FE_HAS_EXTENDED_CAPS, "EXTENDED_CAPS" )
 #if DVBAPI_VERSION >= 501
-    FRONTEND_INFO( info.caps, FE_CAN_2G_MODULATION, "2G_MODULATION" )
+    FRONTEND_INFO( info->caps, FE_CAN_2G_MODULATION, "2G_MODULATION" )
 #endif
-    FRONTEND_INFO( info.caps, FE_NEEDS_BENDING, "NEEDS_BENDING" )
-    FRONTEND_INFO( info.caps, FE_CAN_RECOVER, "FE_CAN_RECOVER" )
-    FRONTEND_INFO( info.caps, FE_CAN_MUTE_TS, "FE_CAN_MUTE_TS" )
+    FRONTEND_INFO( info->caps, FE_CAN_MULTISTREAM, "MULTISTREAM" )
+    FRONTEND_INFO( info->caps, FE_NEEDS_BENDING, "NEEDS_BENDING" )
+    FRONTEND_INFO( info->caps, FE_CAN_RECOVER, "FE_CAN_RECOVER" )
+    FRONTEND_INFO( info->caps, FE_CAN_MUTE_TS, "FE_CAN_MUTE_TS" )
 #undef FRONTEND_INFO
+
+    msg_Dbg( NULL, " delivery systems:" );
+    int i;
+    for ( i = 0; i < i_systems; i++ )
+    {
+        switch ( p_systems[i] )
+        {
+#define DELSYS_INFO(delsys, msg)                                            \
+        case delsys: msg_Dbg( NULL, "  %s", msg); break;
+        DELSYS_INFO( SYS_ATSC, "ATSC" )
+        DELSYS_INFO( SYS_ATSCMH, "ATSCMH" )
+        DELSYS_INFO( SYS_CMMB, "CMBB" )
+        DELSYS_INFO( SYS_DAB, "DAB" )
+        DELSYS_INFO( SYS_DSS, "DSS" )
+        DELSYS_INFO( SYS_DVBC_ANNEX_B, "DVBC_ANNEX_B" )
+        DELSYS_INFO( SYS_DVBH, "DVBH" )
+        DELSYS_INFO( SYS_DVBS, "DVBS" )
+        DELSYS_INFO( SYS_DVBS2, "DVBS2" )
+        DELSYS_INFO( SYS_DVBT, "DVBT" )
+        DELSYS_INFO( SYS_ISDBC, "ISDBC" )
+        DELSYS_INFO( SYS_ISDBS, "ISDBS" )
+        DELSYS_INFO( SYS_ISDBT, "ISDBT" )
+        DELSYS_INFO( SYS_UNDEFINED, "UNDEFINED" )
+#if DVBAPI_VERSION >= 505
+        DELSYS_INFO( SYS_DVBC_ANNEX_A, "DVBC_ANNEX_A" )
+        DELSYS_INFO( SYS_DVBC_ANNEX_C, "DVBC_ANNEX_C" )
+        DELSYS_INFO( SYS_DVBT2, "DVBT2" )
+        DELSYS_INFO( SYS_TURBO, "TURBO" )
+#else
+        DELSYS_INFO( SYS_DVBC_ANNEX_AC, "DVBC_ANNEX_AC" )
+#endif
+#if DVBAPI_VERSION >= 507
+        DELSYS_INFO( SYS_DTMB, "DTMB" )
+#else
+        DELSYS_INFO( SYS_DMBTH, "DMBTH" )
+#endif
+        default: msg_Dbg( NULL, "  Unknown delivery system %u", p_systems[i]);
+            break;
+        }
+    }
 }
 
 /*****************************************************************************
  * FrontendSet
  *****************************************************************************/
 /* S2API */
+#if DVBAPI_VERSION >= 505
+static struct dtv_property info_cmdargs[] = {
+    { .cmd = DTV_API_VERSION,     .u.data = 0 },
+};
+static struct dtv_properties info_cmdseq = {
+    .num = sizeof(info_cmdargs)/sizeof(struct dtv_property),
+    .props = info_cmdargs
+};
+
+static struct dtv_property enum_cmdargs[] = {
+    { .cmd = DTV_ENUM_DELSYS,     .u.data = 0 },
+};
+static struct dtv_properties enum_cmdseq = {
+    .num = sizeof(enum_cmdargs)/sizeof(struct dtv_property),
+    .props = enum_cmdargs
+};
+#endif
+
 static struct dtv_property dvbs_cmdargs[] = {
+    { .cmd = DTV_DELIVERY_SYSTEM, .u.data = SYS_DVBS },
     { .cmd = DTV_FREQUENCY,       .u.data = 0 },
     { .cmd = DTV_MODULATION,      .u.data = QPSK },
     { .cmd = DTV_INVERSION,       .u.data = INVERSION_AUTO },
     { .cmd = DTV_SYMBOL_RATE,     .u.data = 27500000 },
     { .cmd = DTV_INNER_FEC,       .u.data = FEC_AUTO },
-    { .cmd = DTV_DELIVERY_SYSTEM, .u.data = SYS_DVBS },
     { .cmd = DTV_TUNE },
 };
 static struct dtv_properties dvbs_cmdseq = {
@@ -773,14 +894,15 @@ static struct dtv_properties dvbs_cmdseq = {
 };
 
 static struct dtv_property dvbs2_cmdargs[] = {
+    { .cmd = DTV_DELIVERY_SYSTEM, .u.data = SYS_DVBS2 },
     { .cmd = DTV_FREQUENCY,       .u.data = 0 },
     { .cmd = DTV_MODULATION,      .u.data = PSK_8 },
     { .cmd = DTV_INVERSION,       .u.data = INVERSION_AUTO },
     { .cmd = DTV_SYMBOL_RATE,     .u.data = 27500000 },
     { .cmd = DTV_INNER_FEC,       .u.data = FEC_AUTO },
-    { .cmd = DTV_DELIVERY_SYSTEM, .u.data = SYS_DVBS2 },
     { .cmd = DTV_PILOT,           .u.data = PILOT_AUTO },
     { .cmd = DTV_ROLLOFF,         .u.data = ROLLOFF_AUTO },
+    { .cmd = DTV_STREAM_ID,       .u.data = 0 },
     { .cmd = DTV_TUNE },
 };
 static struct dtv_properties dvbs2_cmdseq = {
@@ -789,11 +911,15 @@ static struct dtv_properties dvbs2_cmdseq = {
 };
 
 static struct dtv_property dvbc_cmdargs[] = {
+#if DVBAPI_VERSION >= 505
+    { .cmd = DTV_DELIVERY_SYSTEM, .u.data = SYS_DVBC_ANNEX_A },
+#else
+    { .cmd = DTV_DELIVERY_SYSTEM, .u.data = SYS_DVBC_ANNEX_AC },
+#endif
     { .cmd = DTV_FREQUENCY,       .u.data = 0 },
     { .cmd = DTV_MODULATION,      .u.data = QAM_AUTO },
     { .cmd = DTV_INVERSION,       .u.data = INVERSION_AUTO },
     { .cmd = DTV_SYMBOL_RATE,     .u.data = 27500000 },
-    { .cmd = DTV_DELIVERY_SYSTEM, .u.data = SYS_DVBC_ANNEX_AC },
     { .cmd = DTV_TUNE },
 };
 static struct dtv_properties dvbc_cmdseq = {
@@ -802,6 +928,7 @@ static struct dtv_properties dvbc_cmdseq = {
 };
 
 static struct dtv_property dvbt_cmdargs[] = {
+    { .cmd = DTV_DELIVERY_SYSTEM, .u.data = SYS_DVBT },
     { .cmd = DTV_FREQUENCY,       .u.data = 0 },
     { .cmd = DTV_MODULATION,      .u.data = QAM_AUTO },
     { .cmd = DTV_INVERSION,       .u.data = INVERSION_AUTO },
@@ -811,7 +938,6 @@ static struct dtv_property dvbt_cmdargs[] = {
     { .cmd = DTV_GUARD_INTERVAL,  .u.data = GUARD_INTERVAL_AUTO },
     { .cmd = DTV_TRANSMISSION_MODE,.u.data = TRANSMISSION_MODE_AUTO },
     { .cmd = DTV_HIERARCHY,       .u.data = HIERARCHY_AUTO },
-    { .cmd = DTV_DELIVERY_SYSTEM, .u.data = SYS_DVBT },
     { .cmd = DTV_TUNE },
 };
 static struct dtv_properties dvbt_cmdseq = {
@@ -819,11 +945,12 @@ static struct dtv_properties dvbt_cmdseq = {
     .props = dvbt_cmdargs
 };
 
+/* ATSC + DVB-C annex B */
 static struct dtv_property atsc_cmdargs[] = {
-    { .cmd = DTV_FREQUENCY, .u.data = 0 },
-    { .cmd = DTV_MODULATION, .u.data = QAM_AUTO },
-    { .cmd = DTV_INVERSION, .u.data = INVERSION_AUTO },
     { .cmd = DTV_DELIVERY_SYSTEM, .u.data = SYS_ATSC },
+    { .cmd = DTV_FREQUENCY,       .u.data = 0 },
+    { .cmd = DTV_MODULATION,      .u.data = QAM_AUTO },
+    { .cmd = DTV_INVERSION,       .u.data = INVERSION_AUTO },
     { .cmd = DTV_TUNE },
 };
 static struct dtv_properties atsc_cmdseq = {
@@ -831,18 +958,20 @@ static struct dtv_properties atsc_cmdseq = {
     .props = atsc_cmdargs
 };
 
-#define FREQUENCY 0
-#define MODULATION 1
-#define INVERSION 2
-#define SYMBOL_RATE 3
-#define BANDWIDTH 3
-#define FEC_INNER 4
-#define FEC_LP 5
-#define GUARD 6
-#define PILOT 6
-#define TRANSMISSION 7
-#define ROLLOFF 7
-#define HIERARCHY 8
+#define DELSYS 0
+#define FREQUENCY 1
+#define MODULATION 2
+#define INVERSION 3
+#define SYMBOL_RATE 4
+#define BANDWIDTH 4
+#define FEC_INNER 5
+#define FEC_LP 6
+#define GUARD 7
+#define PILOT 7
+#define TRANSMISSION 8
+#define ROLLOFF 8
+#define MIS 9
+#define HIERARCHY 9
 
 struct dtv_property pclear[] = {
     { .cmd = DTV_CLEAR },
@@ -853,10 +982,75 @@ struct dtv_properties cmdclear = {
     .props = pclear
 };
 
+static fe_delivery_system_t
+FrontendGuessSystem( fe_delivery_system_t *p_systems, int i_systems )
+{
+    if ( psz_delsys != NULL )
+    {
+        if ( !strcasecmp( psz_delsys, "DVBS" ) )
+            return SYS_DVBS;
+        if ( !strcasecmp( psz_delsys, "DVBS2" ) )
+            return SYS_DVBS2;
+        if ( !strcasecmp( psz_delsys, "DVBC_ANNEX_A" ) )
+#if DVBAPI_VERSION >= 505
+            return SYS_DVBC_ANNEX_A;
+#else
+            return SYS_DVBC_ANNEX_AC;
+#endif
+        if ( !strcasecmp( psz_delsys, "DVBC_ANNEX_B" ) )
+            return SYS_DVBC_ANNEX_B;
+        if ( !strcasecmp( psz_delsys, "DVBT" ) )
+            return SYS_DVBT;
+        if ( !strcasecmp( psz_delsys, "ATSC" ) )
+            return SYS_ATSC;
+        msg_Err( NULL, "unknown delivery system %s", psz_delsys );
+        exit(1);
+    }
+
+    if ( i_systems == 1 )
+        return p_systems[0];
+
+    int i;
+    for ( i = 0; i < i_systems; i++ )
+    {
+        switch ( p_systems[i] )
+        {
+            case SYS_DVBS:
+                if ( i_frequency < 50000000 )
+                    return SYS_DVBS;
+                break;
+#if DVBAPI_VERSION >= 505
+            case SYS_DVBC_ANNEX_A:
+                if ( i_frequency > 50000000 || i_srate != 27500000 ||
+                     psz_modulation != NULL )
+                    return SYS_DVBC_ANNEX_A;
+                break;
+#else
+            case SYS_DVBC_ANNEX_AC:
+                if ( i_frequency > 50000000 || i_srate != 27500000 ||
+                     psz_modulation != NULL )
+                    return SYS_DVBC_ANNEX_AC;
+                break;
+#endif
+            case SYS_DVBT:
+                if ( i_frequency > 50000000 )
+                    return SYS_DVBT;
+                break;
+            default:
+                break;
+        }
+    }
+
+    msg_Warn( NULL, "couldn't guess delivery system, use --delsys" );
+    return p_systems[0];
+}
+
 static void FrontendSet( bool b_init )
 {
     struct dvb_frontend_info info;
     struct dtv_properties *p;
+    fe_delivery_system_t p_systems[MAX_DELIVERY_SYSTEMS] = { 0 };
+    int i_systems = 0;
 
     if ( ioctl( i_frontend, FE_GET_INFO, &info ) < 0 )
     {
@@ -864,8 +1058,68 @@ static void FrontendSet( bool b_init )
         exit(1);
     }
 
+    uint32_t version = 0x300;
+#if DVBAPI_VERSION >= 505
+    if ( ioctl( i_frontend, FE_GET_PROPERTY, &info_cmdseq ) < 0 )
+    {
+#endif
+        /* DVBv3 device */
+        switch ( info.type )
+        {
+        case FE_OFDM:
+            p_systems[i_systems++] = SYS_DVBT;
+#if DVBAPI_VERSION >= 505
+            if ( info.caps & FE_CAN_2G_MODULATION )
+                p_systems[i_systems++] = SYS_DVBT2;
+#endif
+            break;
+        case FE_QAM:
+#if DVBAPI_VERSION >= 505
+            p_systems[i_systems++] = SYS_DVBC_ANNEX_A;
+#else
+            p_systems[i_systems++] = SYS_DVBC_ANNEX_AC;
+#endif
+            break;
+        case FE_QPSK:
+            p_systems[i_systems++] = SYS_DVBS;
+            if ( info.caps & FE_CAN_2G_MODULATION )
+                p_systems[i_systems++] = SYS_DVBS2;
+            break;
+        case FE_ATSC:
+            if ( info.caps & (FE_CAN_8VSB | FE_CAN_16VSB) )
+                p_systems[i_systems++] = SYS_ATSC;
+            if ( info.caps & (FE_CAN_QAM_64 | FE_CAN_QAM_256 | FE_CAN_QAM_AUTO) )
+                p_systems[i_systems++] = SYS_DVBC_ANNEX_B;
+            break;
+        default:
+            msg_Err( NULL, "unknown frontend type %d", info.type );
+            exit(1);
+        }
+#if DVBAPI_VERSION >= 505
+    }
+    else
+    {
+        version = info_cmdargs[0].u.data;
+        if ( ioctl( i_frontend, FE_GET_PROPERTY, &enum_cmdseq ) < 0 )
+        {
+            msg_Err( NULL, "unable to query frontend" );
+            exit(1);
+        }
+        i_systems = enum_cmdargs[0].u.buffer.len;
+        if ( i_systems < 1 )
+        {
+            msg_Err( NULL, "no available delivery system" );
+            exit(1);
+        }
+
+        int i;
+        for ( i = 0; i < i_systems; i++ )
+            p_systems[i] = enum_cmdargs[0].u.buffer.data[i];
+    }
+#endif
+
     if ( b_init )
-        FrontendInfo( info );
+        FrontendInfo( &info, version, p_systems, i_systems );
 
     /* Clear frontend commands */
     if ( ioctl( i_frontend, FE_SET_PROPERTY, &cmdclear ) < 0 )
@@ -874,10 +1128,12 @@ static void FrontendSet( bool b_init )
         exit(1);
     }
 
-    switch ( info.type )
+    fe_delivery_system_t system = FrontendGuessSystem( p_systems, i_systems );
+    switch ( system )
     {
-    case FE_OFDM:
+    case SYS_DVBT:
         p = &dvbt_cmdseq;
+        p->props[DELSYS].u.data = system;
         p->props[FREQUENCY].u.data = i_frequency;
         p->props[INVERSION].u.data = GetInversion();
         if ( psz_modulation != NULL )
@@ -889,14 +1145,18 @@ static void FrontendSet( bool b_init )
         p->props[TRANSMISSION].u.data = GetTransmission();
         p->props[HIERARCHY].u.data = GetHierarchy();
 
-        msg_Dbg( NULL, "tuning OFDM frontend to f=%d bandwidth=%d inversion=%d fec_hp=%d fec_lp=%d hierarchy=%d modulation=%s guard=%d transmission=%d",
+        msg_Dbg( NULL, "tuning DVB-T frontend to f=%d bandwidth=%d inversion=%d fec_hp=%d fec_lp=%d hierarchy=%d modulation=%s guard=%d transmission=%d",
                  i_frequency, i_bandwidth, i_inversion, i_fec, i_fec_lp,
                  i_hierarchy,
                  psz_modulation == NULL ? "qam_auto" : psz_modulation,
                  i_guard, i_transmission );
         break;
 
-    case FE_QAM:
+#if DVBAPI_VERSION >= 505
+    case SYS_DVBC_ANNEX_A:
+#else
+    case SYS_DVBC_ANNEX_AC:
+#endif
         p = &dvbc_cmdseq;
         p->props[FREQUENCY].u.data = i_frequency;
         p->props[INVERSION].u.data = GetInversion();
@@ -904,18 +1164,33 @@ static void FrontendSet( bool b_init )
             p->props[MODULATION].u.data = GetModulation();
         p->props[SYMBOL_RATE].u.data = i_srate;
 
-        msg_Dbg( NULL, "tuning QAM frontend to f=%d srate=%d inversion=%d modulation=%s",
+        msg_Dbg( NULL, "tuning DVB-C frontend to f=%d srate=%d inversion=%d modulation=%s",
                  i_frequency, i_srate, i_inversion,
                  psz_modulation == NULL ? "qam_auto" : psz_modulation );
         break;
 
-    case FE_QPSK:
+    case SYS_DVBC_ANNEX_B:
+        p = &atsc_cmdseq;
+        p->props[DELSYS].u.data = system;
+        p->props[FREQUENCY].u.data = i_frequency;
+        p->props[INVERSION].u.data = GetInversion();
+        if ( psz_modulation != NULL )
+            p->props[MODULATION].u.data = GetModulation();
+
+        msg_Dbg( NULL, "tuning ATSC cable frontend to f=%d inversion=%d modulation=%s",
+                 i_frequency, i_inversion,
+                 psz_modulation == NULL ? "qam_auto" : psz_modulation );
+        break;
+
+    case SYS_DVBS:
+    case SYS_DVBS2:
         if ( psz_modulation != NULL )
         {
             p = &dvbs2_cmdseq;
             p->props[MODULATION].u.data = GetModulation();
             p->props[PILOT].u.data = GetPilot();
             p->props[ROLLOFF].u.data = GetRollOff();
+            p->props[MIS].u.data = i_mis;
         }
         else
             p = &dvbs_cmdseq;
@@ -925,12 +1200,13 @@ static void FrontendSet( bool b_init )
         p->props[FEC_INNER].u.data = GetFECInner(info.caps);
         p->props[FREQUENCY].u.data = FrontendDoDiseqc();
 
-        msg_Dbg( NULL, "tuning QPSK frontend to f=%d srate=%d inversion=%d fec=%d rolloff=%d modulation=%s pilot=%d",
+        msg_Dbg( NULL, "tuning DVB-S frontend to f=%d srate=%d inversion=%d fec=%d rolloff=%d modulation=%s pilot=%d mis=%d",
                  i_frequency, i_srate, i_inversion, i_fec, i_rolloff,
-                 psz_modulation == NULL ? "legacy" : psz_modulation, i_pilot );
+                 psz_modulation == NULL ? "legacy" : psz_modulation, i_pilot,
+                 i_mis );
         break;
 
-    case FE_ATSC:
+    case SYS_ATSC:
         p = &atsc_cmdseq;
         p->props[FREQUENCY].u.data = i_frequency;
         p->props[INVERSION].u.data = GetInversion();
@@ -964,7 +1240,9 @@ static void FrontendSet( bool b_init )
     }
 
     i_last_status = 0;
-    i_frontend_timeout = i_wallclock + i_frontend_timeout_duration;
+
+    if (i_frontend_timeout_duration)
+        ev_timer_again(event_loop, &lock_watcher);
 }
 
 #else /* !S2API */
@@ -1061,7 +1339,9 @@ static void FrontendSet( bool b_init )
     }
 
     i_last_status = 0;
-    i_frontend_timeout = i_wallclock + i_frontend_timeout_duration;
+
+    if (i_frontend_timeout_duration)
+        ev_timer_again(event_loop, &lock_watcher);
 }
 
 #endif /* S2API */
@@ -1102,3 +1382,5 @@ uint8_t dvb_FrontendStatus( uint8_t *p_answer, ssize_t *pi_size )
     *pi_size = sizeof(struct ret_frontend_status);
     return RET_FRONTEND_STATUS;
 }
+
+#endif
