@@ -1,7 +1,7 @@
 /*****************************************************************************
  * dvblast.c
  *****************************************************************************
- * Copyright (C) 2004, 2008-2011 VideoLAN
+ * Copyright (C) 2004, 2008-2011, 2015 VideoLAN
  *
  * Authors: Christophe Massiot <massiot@via.ecp.fr>
  *          Andy Gatward <a.j.gatward@reading.ac.uk>
@@ -38,6 +38,8 @@
 #include <getopt.h>
 #include <errno.h>
 
+#include <ev.h>
+
 #include "dvblast.h"
 
 #ifdef HAVE_ICONV
@@ -52,49 +54,50 @@
 /*****************************************************************************
  * Local declarations
  *****************************************************************************/
-mtime_t i_wallclock = 0;
+struct ev_loop *event_loop;
 output_t **pp_outputs = NULL;
 int i_nb_outputs = 0;
 output_t output_dup;
-static char *psz_conf_file = NULL;
+bool b_passthrough = false;
+static const char *psz_conf_file = NULL;
 char *psz_srv_socket = NULL;
 static int i_priority = -1;
 int i_adapter = 0;
 int i_fenum = 0;
+int i_canum = 0;
+char *psz_delsys = NULL;
 int i_frequency = 0;
 int i_inversion = -1;
 int i_srate = 27500000;
 int i_fec = 999;
 int i_rolloff = 35;
 int i_satnum = 0;
+int i_uncommitted = 0;
 int i_voltage = 13;
 int b_tone = 0;
 int i_bandwidth = 8;
 char *psz_modulation = NULL;
 int i_pilot = -1;
+int i_mis = 0;
 int i_fec_lp = 999;
 int i_guard = -1;
 int i_transmission = -1;
 int i_hierarchy = -1;
 mtime_t i_frontend_timeout_duration = DEFAULT_FRONTEND_TIMEOUT;
-mtime_t i_quit_timeout = 0;
 mtime_t i_quit_timeout_duration = 0;
 int b_budget_mode = 0;
 int b_any_type = 0;
 int b_select_pmts = 0;
 int b_random_tsid = 0;
-uint16_t i_network_id = 0xffff;
-uint8_t *p_network_name;
-size_t i_network_name_size;
 char *psz_udp_src = NULL;
 int i_asi_adapter = 0;
 const char *psz_native_charset = "UTF-8";
-const char *psz_dvb_charset = "ISO_8859-1";
-const char *psz_provider_name = NULL;
-print_type_t i_print_type = -1;
+print_type_t i_print_type = PRINT_TEXT;
+bool b_print_enabled = false;
+FILE *print_fh;
+mtime_t i_print_period = 0;
+mtime_t i_es_timeout = 0;
 
-volatile sig_atomic_t b_conf_reload = 0;
-volatile sig_atomic_t b_exit_now = 0;
 int i_verbose = DEFAULT_VERBOSITY;
 int i_syslog = 0;
 char *psz_syslog_ident = NULL;
@@ -110,11 +113,20 @@ static mtime_t i_latency_global = DEFAULT_OUTPUT_LATENCY;
 static mtime_t i_retention_global = DEFAULT_MAX_RETENTION;
 static int i_ttl_global = 64;
 
+static const char *psz_dvb_charset = "UTF-8";
+static iconv_t conf_iconv = (iconv_t)-1;
+static uint16_t i_network_id = 0xffff;
+static dvb_string_t network_name;
+static dvb_string_t provider_name;
+
 /* TPS Input log filename */
 char * psz_mrtg_file = NULL;
 
+/* PID mapping */
+bool b_do_remap = false;
+uint16_t pi_newpids[ N_MAP_PIDS ];  /* pmt, audio, video, spu */
+
 void (*pf_Open)( void ) = NULL;
-block_t * (*pf_Read)( mtime_t i_poll_timeout ) = NULL;
 void (*pf_Reset)( void ) = NULL;
 int (*pf_SetFilter)( uint16_t i_pid ) = NULL;
 void (*pf_UnsetFilter)( int i_fd, uint16_t i_pid ) = NULL;
@@ -127,23 +139,35 @@ void config_Init( output_config_t *p_config )
     memset( p_config, 0, sizeof(output_config_t) );
 
     p_config->psz_displayname = NULL;
-    p_config->psz_service_name = NULL;
-    p_config->psz_service_provider = NULL;
+    p_config->i_network_id = i_network_id;
+    dvb_string_init(&p_config->network_name);
+    dvb_string_init(&p_config->service_name);
+    dvb_string_init(&p_config->provider_name);
+    p_config->psz_srcaddr = NULL;
 
     p_config->i_family = AF_UNSPEC;
     p_config->connect_addr.ss_family = AF_UNSPEC;
     p_config->bind_addr.ss_family = AF_UNSPEC;
     p_config->i_if_index_v6 = -1;
+    p_config->i_srcport = 0;
 
     p_config->pi_pids = NULL;
+    p_config->b_passthrough = false;
+    p_config->b_do_remap = false;
+    unsigned int i;
+    for ( i = 0; i < N_MAP_PIDS; i++ ) {
+        p_config->pi_confpids[i]  = UNUSED_PID;
+    }
 }
 
 void config_Free( output_config_t *p_config )
 {
     free( p_config->psz_displayname );
-    free( p_config->psz_service_name );
-    free( p_config->psz_service_provider );
+    dvb_string_clean( &p_config->network_name );
+    dvb_string_clean( &p_config->service_name );
+    dvb_string_clean( &p_config->provider_name );
     free( p_config->pi_pids );
+    free( p_config->psz_srcaddr );
 }
 
 static void config_Defaults( output_config_t *p_config )
@@ -158,9 +182,11 @@ static void config_Defaults( output_config_t *p_config )
     p_config->i_tsid = -1;
     p_config->i_ttl = i_ttl_global;
     memcpy( p_config->pi_ssrc, pi_ssrc_global, 4 * sizeof(uint8_t) );
+    dvb_string_copy(&p_config->network_name, &network_name);
+    dvb_string_copy(&p_config->provider_name, &provider_name);
 }
 
-static char *config_stropt( char *psz_string )
+char *config_stropt( const char *psz_string )
 {
     char *ret, *tmp;
     if ( !psz_string || strlen( psz_string ) == 0 )
@@ -178,7 +204,56 @@ static char *config_stropt( char *psz_string )
     return ret;
 }
 
-bool config_ParseHost( output_config_t *p_config, char *psz_string )
+static char *config_striconv( const char *psz_string )
+{
+    char *psz_input = config_stropt(psz_string);
+
+    if ( !strcasecmp( psz_native_charset, psz_dvb_charset ) )
+        return psz_input;
+
+#ifdef HAVE_ICONV
+    if ( conf_iconv == (iconv_t)-1 )
+    {
+        conf_iconv = iconv_open( psz_dvb_charset, psz_native_charset );
+        if ( conf_iconv == (iconv_t)-1 )
+            return psz_input;
+    }
+
+    size_t i_input = strlen( psz_input );
+    size_t i_output = i_input * 6;
+    char *psz_output = malloc( i_output );
+    char *p = psz_output;
+    if ( iconv( conf_iconv, &psz_input, &i_input, &p, &i_output ) == -1 )
+    {
+        free( psz_output );
+        return psz_input;
+    }
+
+    free(psz_input);
+    return psz_output;
+#else
+    msg_Warn( NULL,
+              "unable to convert from %s to %s (iconv is not available)",
+              psz_native_charset, psz_dvb_charset );
+    return psz_input;
+#endif
+}
+
+static void config_strdvb( dvb_string_t *p_dvb_string, const char *psz_string )
+{
+    if (psz_string == NULL)
+    {
+        dvb_string_init(p_dvb_string);
+        return;
+    }
+
+    char *psz_iconv = config_striconv(psz_string);
+    dvb_string_clean(p_dvb_string);
+    p_dvb_string->p = dvb_string_set((uint8_t *)psz_iconv, strlen(psz_iconv),
+                                     psz_dvb_charset, &p_dvb_string->i);
+}
+
+static bool config_ParseHost( output_config_t *p_config, char *psz_string )
 {
     struct addrinfo *p_ai;
     int i_mtu;
@@ -235,23 +310,55 @@ bool config_ParseHost( output_config_t *p_config, char *psz_string )
             p_config->i_mtu = strtol( ARG_OPTION("mtu="), NULL, 0 );
         else if ( IS_OPTION("ifindex=") )
             p_config->i_if_index_v6 = strtol( ARG_OPTION("ifindex="), NULL, 0 );
+        else if ( IS_OPTION("networkid=") )
+            p_config->i_network_id = strtol( ARG_OPTION("networkid="), NULL, 0 );
+        else if ( IS_OPTION("networkname=")  )
+        {
+            config_strdvb( &p_config->network_name, ARG_OPTION("networkname=") );
+        }
         else if ( IS_OPTION("srvname=")  )
         {
-            if ( p_config->psz_service_name )
-                free( p_config->psz_service_name );
-            p_config->psz_service_name = config_stropt( ARG_OPTION("srvname=") );
+            config_strdvb( &p_config->service_name, ARG_OPTION("srvname=") );
         }
         else if ( IS_OPTION("srvprovider=") )
         {
-            if ( !p_config->psz_service_provider )
-                free( p_config->psz_service_provider );
-            p_config->psz_service_provider = config_stropt( ARG_OPTION("srvprovider=") );
+            config_strdvb( &p_config->provider_name, ARG_OPTION("srvprovider=") );
         }
+        else if ( IS_OPTION("srcaddr=") )
+        {
+            if ( p_config->i_family != AF_INET ) {
+                msg_Err( NULL, "RAW sockets currently implemented for ipv4 only");
+                return false;
+            }
+            free( p_config->psz_srcaddr );
+            p_config->psz_srcaddr = config_stropt( ARG_OPTION("srcaddr=") );
+            p_config->i_config |= OUTPUT_RAW;
+        }
+        else if ( IS_OPTION("srcport=") )
+            p_config->i_srcport = strtol( ARG_OPTION("srcport="), NULL, 0 );
         else if ( IS_OPTION("ssrc=") )
         {
             in_addr_t i_addr = inet_addr( ARG_OPTION("ssrc=") );
             memcpy( p_config->pi_ssrc, &i_addr, 4 * sizeof(uint8_t) );
         }
+        else if ( IS_OPTION("pidmap=") )
+        {
+            char *str1;
+            char *saveptr = NULL;
+            char *tok = NULL;
+            int i, i_newpid;
+            for (i = 0, str1 = config_stropt( (ARG_OPTION("pidmap="))); i < N_MAP_PIDS; i++, str1 = NULL)
+            {
+                tok = strtok_r(str1, ",", &saveptr);
+                if ( !tok )
+                    break;
+                i_newpid = strtoul(tok, NULL, 0);
+                p_config->pi_confpids[i] = i_newpid;
+            }
+            p_config->b_do_remap = true;
+        }
+        else if ( IS_OPTION("newsid=") )
+            p_config->i_new_sid = strtol( ARG_OPTION("newsid="), NULL, 0 );
         else
             msg_Warn( NULL, "unrecognized option %s", psz_string );
 
@@ -259,9 +366,6 @@ bool config_ParseHost( output_config_t *p_config, char *psz_string )
 #undef ARG_OPTION
     }
 
-    if ( !p_config->psz_service_provider && psz_provider_name )
-        p_config->psz_service_provider = strdup( psz_provider_name );
-
 end:
     i_mtu = p_config->i_family == AF_INET6 ? DEFAULT_IPV6_MTU :
             DEFAULT_IPV4_MTU;
@@ -279,6 +383,13 @@ end:
 
 static void config_Print( output_config_t *p_config )
 {
+    if ( p_config->b_passthrough )
+    {
+        msg_Dbg( NULL, "conf: %s config=0x%"PRIx64" sid=*",
+                 p_config->psz_displayname, p_config->i_config);
+        return;
+    }
+
     const char *psz_base = "conf: %s config=0x%"PRIx64" sid=%hu pids[%d]=";
     size_t i_len = strlen(psz_base) + 6 * p_config->i_nb_pids + 1;
     char psz_format[i_len];
@@ -293,21 +404,21 @@ static void config_Print( output_config_t *p_config )
              p_config->i_sid, p_config->i_nb_pids );
 }
 
-static void config_ReadFile( char *psz_file )
+void config_ReadFile(void)
 {
     FILE *p_file;
     char psz_line[2048];
     int i;
 
-    if ( psz_file == NULL )
+    if ( psz_conf_file == NULL )
     {
         msg_Err( NULL, "no config file" );
         return;
     }
 
-    if ( (p_file = fopen( psz_file, "r" )) == NULL )
+    if ( (p_file = fopen( psz_conf_file, "r" )) == NULL )
     {
-        msg_Err( NULL, "can't fopen config file %s", psz_file );
+        msg_Err( NULL, "can't fopen config file %s", psz_conf_file );
         return;
     }
 
@@ -351,21 +462,29 @@ static void config_ReadFile( char *psz_file )
             config_Free( &config );
             continue;
         }
-        config.i_sid = strtol(psz_token, NULL, 0);
 
-        psz_token = strtok_r( NULL, "\t\n ", &psz_parser );
-        if ( psz_token != NULL )
+        if ( psz_token[0] == '*' )
         {
-            psz_parser = NULL;
-            for ( ; ; )
+            config.b_passthrough = true;
+        }
+        else
+        {
+            config.i_sid = strtol(psz_token, NULL, 0);
+
+            psz_token = strtok_r( NULL, "\t\n ", &psz_parser );
+            if ( psz_token != NULL )
             {
-                psz_token = strtok_r( psz_token, ",", &psz_parser );
-                if ( psz_token == NULL )
-                    break;
-                config.pi_pids = realloc( config.pi_pids,
-                                 (config.i_nb_pids + 1) * sizeof(uint16_t) );
-                config.pi_pids[config.i_nb_pids++] = strtol(psz_token, NULL, 0);
-                psz_token = NULL;
+                psz_parser = NULL;
+                for ( ; ; )
+                {
+                    psz_token = strtok_r( psz_token, ",", &psz_parser );
+                    if ( psz_token == NULL )
+                        break;
+                    config.pi_pids = realloc( config.pi_pids,
+                                     (config.i_nb_pids + 1) * sizeof(uint16_t) );
+                    config.pi_pids[config.i_nb_pids++] = strtol(psz_token, NULL, 0);
+                    psz_token = NULL;
+                }
             }
         }
 
@@ -414,12 +533,39 @@ static void config_ReadFile( char *psz_file )
 /*****************************************************************************
  * Signal Handler
  *****************************************************************************/
-static void SigHandler( int i_signal )
+static void signal_watcher_init(struct ev_signal *w, struct ev_loop *loop,
+                    void (*cb)(struct ev_loop*, struct ev_signal*, int),
+                    int signum)
 {
-    if ( i_signal == SIGHUP )
-        b_conf_reload = 1;
-    if ( i_signal == SIGINT )
-        b_exit_now = 1;
+    ev_signal_init(w, cb, signum);
+    ev_signal_start(loop, w);
+    ev_unref(loop);
+}
+
+static void sighandler(struct ev_loop *loop, struct ev_signal *w, int revents)
+{
+    switch (w->signum)
+    {
+        case SIGINT:
+        case SIGTERM:
+        default:
+            msg_Info( NULL, "Shutdown was requested." );
+            ev_break(loop, EVBREAK_ALL);
+            break;
+
+        case SIGHUP:
+            msg_Info( NULL, "Configuration reload was requested." );
+            config_ReadFile();
+            break;
+    }
+}
+
+/*****************************************************************************
+ * Quit timeout
+ *****************************************************************************/
+static void quit_cb(struct ev_loop *loop, struct ev_timer *w, int revents)
+{
+    ev_break(loop, EVBREAK_ALL);
 }
 
 /*****************************************************************************
@@ -437,21 +583,37 @@ void usage()
 {
     DisplayVersion();
     msg_Raw( NULL, "Usage: dvblast [-q] [-c <config file>] [-r <remote socket>] [-t <ttl>] [-o <SSRC IP>] "
-        "[-i <RT priority>] [-a <adapter>] [-n <frontend number>] [-S <diseqc>] "
-        "[-f <frequency>|-D [<src host>[:<src port>]@]<src mcast>[:<port>][/<opts>]*|-A <ASI adapter>] "
+        "[-i <RT priority>] "
+#ifdef HAVE_ASI_SUPPORT
+        "[-A <ASI adapter>]"
+#endif
+#ifdef HAVE_DVB_SUPPORT
+        "[-a <adapter>] [-n <frontend number>] [-S <diseqc>] [-k <uncommitted port>]"
+        "[-f <frequency>]"
         "[-s <symbol rate>] [-v <0|13|18>] [-p] [-b <bandwidth>] [-I <inversion>] "
         "[-F <fec inner>] [-m <modulation] [-R <rolloff>] [-P <pilot>] [-K <fec lp>] "
         "[-G <guard interval>] [-H <hierarchy>] [-X <transmission>] [-O <lock timeout>] "
-        "[-u] [-w] [-U] [-L <latency>] [-E <retention>] [-d <dest IP>[<:port>][/<opts>]*] "
-        "[-z] [-C [-e] [-M <network name] [-N <network ID>]] [-T] [-j <system charset>] "
-        "[-W] [-Y] [-l] [-g <logger ident>] [-Z <mrtg file>] [-V] [-h] [-B <provider_name>]"
-        "[-J <DVB charset>] [-Q <quit timeout>] [-x <text|xml>]" );
+#endif
+        "[-D [<src host>[:<src port>]@]<src mcast>[:<port>][/<opts>]*] "
+        "[-u] [-w] [-U] [-L <latency>] [-E <retention>] [-d <dest IP>[<:port>][/<opts>]*] [-3] "
+        "[-z] [-C [-e] [-M <network name>] [-N <network ID>]] [-T] [-j <system charset>] "
+        "[-W] [-Y] [-l] [-g <logger ident>] [-Z <mrtg file>] [-V] [-h] [-B <provider_name>] "
+        "[-1 <mis_id>] [-2 <size>] [-5 <DVBS|DVBS2|DVBC_ANNEX_A|DVBT|ATSC>] -y <ca_dev_number> "
+        "[-J <DVB charset>] [-Q <quit timeout>] [-0 pid_mapping] [-x <text|xml>]"
+        "[-6 <print period>] [-7 <ES timeout>]" );
 
     msg_Raw( NULL, "Input:" );
-    msg_Raw( NULL, "  -a --adapter          read packets from a Linux-DVB adapter (typically 0-n)" );
+#ifdef HAVE_ASI_SUPPORT
     msg_Raw( NULL, "  -A --asi-adapter      read packets from an ASI adapter (0-n)" );
+#endif
+#ifdef HAVE_DVB_SUPPORT
+    msg_Raw( NULL, "  -a --adapter          read packets from a Linux-DVB adapter (typically 0-n)" );
     msg_Raw( NULL, "  -b --bandwidth        frontend bandwith" );
+#endif
     msg_Raw( NULL, "  -D --rtp-input        read packets from a multicast address instead of a DVB card" );
+#ifdef HAVE_DVB_SUPPORT
+    msg_Raw( NULL, "  -5 --delsys           delivery system" );
+    msg_Raw( NULL, "    DVBS|DVBS2|DVBC_ANNEX_A|DVBT|ATSC (default guessed)");
     msg_Raw( NULL, "  -f --frequency        frontend frequency" );
     msg_Raw( NULL, "  -F --fec-inner        Forward Error Correction (FEC Inner)");
     msg_Raw( NULL, "    DVB-S2 0|12|23|34|35|56|78|89|910|999 (default auto: 999)");
@@ -465,6 +627,7 @@ void usage()
     msg_Raw( NULL, "  -P --pilot            DVB-S2 Pilot (-1 auto, 0 off, 1 on)" );
     msg_Raw( NULL, "  -R --rolloff          DVB-S2 Rolloff value" );
     msg_Raw( NULL, "    DVB-S2 35=0.35|25=0.25|20=0.20|0=AUTO (default: 35)" );
+    msg_Raw( NULL, "  -1 --multistream-id   Set stream ID (0-255, default: 0)" );
     msg_Raw( NULL, "  -K --fec-lp           DVB-T low priority FEC (default auto)" );
     msg_Raw( NULL, "  -G --guard            DVB-T guard interval" );
     msg_Raw( NULL, "    DVB-T  32 (1/32)|16 (1/16)|8 (1/8)|4 (1/4)|-1 (auto, default)" );
@@ -472,15 +635,20 @@ void usage()
     msg_Raw( NULL, "  -X --transmission     DVB-T transmission (2, 4, 8 or -1 auto, default)" );
     msg_Raw( NULL, "  -s --symbol-rate" );
     msg_Raw( NULL, "  -S --diseqc           satellite number for diseqc (0: no diseqc, 1-4, A or B)" );
+    msg_Raw( NULL, "  -k --uncommitted      port number for uncommitted diseqc (0: no uncommitted diseqc, 1-4)" );
     msg_Raw( NULL, "  -u --budget-mode      turn on budget mode (no hardware PID filtering)" );
     msg_Raw( NULL, "  -v --voltage          voltage to apply to the LNB (QPSK)" );
-    msg_Raw( NULL, "  -w --select-pmts      set a PID filter on all PMTs" );
+    msg_Raw( NULL, "  -w --select-pmts      set a PID filter on all PMTs (auto on, when config file is used)" );
     msg_Raw( NULL, "  -O --lock-timeout     timeout for the lock operation (in ms)" );
+    msg_Raw( NULL, "  -y --ca-number <ca_device_number>" );
+    msg_Raw( NULL, "  -2 --dvr-buf-size <size> set the size of the DVR TS buffer in bytes (default: %d)", i_dvr_buffer_size);
+#endif
 
     msg_Raw( NULL, "Output:" );
     msg_Raw( NULL, "  -c --config-file <config file>" );
     msg_Raw( NULL, "  -C --dvb-compliance   pass through or build the mandatory DVB tables" );
     msg_Raw( NULL, "  -d --duplicate        duplicate all received packets to a given destination" );
+    msg_Raw( NULL, "  -3 --passthrough      duplicate all received packets to stdout" );
     msg_Raw( NULL, "  -W --emm-passthrough  pass through EMM data (CA system data)" );
     msg_Raw( NULL, "  -Y --ecm-passthrough  pass through ECM data (CA program data)" );
     msg_Raw( NULL, "  -e --epg-passthrough  pass through DVB EIT schedule tables" );
@@ -494,43 +662,45 @@ void usage()
     msg_Raw( NULL, "  -T --unique-ts-id     generate random unique TS ID for each output" );
     msg_Raw( NULL, "  -U --udp              use raw UDP rather than RTP (required by some IPTV set top boxes)" );
     msg_Raw( NULL, "  -z --any-type         pass through all ESs from the PMT, of any type" );
+    msg_Raw( NULL, "  -0 --pidmap <pmt_pid,audio_pid,video_pid,spu_pid>");
 
     msg_Raw( NULL, "Misc:" );
     msg_Raw( NULL, "  -h --help             display this full help" );
     msg_Raw( NULL, "  -i --priority <RT priority>" );
     msg_Raw( NULL, "  -j --system-charset   character set used for printing messages (default UTF-8)" );
-    msg_Raw( NULL, "  -J --dvb-charset      character set used in output DVB tables (default ISO_8859-1)" );
+    msg_Raw( NULL, "  -J --dvb-charset      character set used in output DVB tables (default UTF-8)" );
     msg_Raw( NULL, "  -l --logger           use syslog for logging messages instead of stderr" );
     msg_Raw( NULL, "  -g --logger-ident     program name that will be used in syslog messages" );
     msg_Raw( NULL, "  -x --print            print interesting events on stdout in a given format" );
     msg_Raw( NULL, "  -q --quiet            be quiet (less verbosity, repeat or use number for even quieter)" );
     msg_Raw( NULL, "  -Q --quit-timeout     when locked, quit after this delay (in ms), or after the first lock timeout" );
+    msg_Raw( NULL, "  -6 --print-period     periodicity at which we print bitrate and errors (in ms)" );
+    msg_Raw( NULL, "  -7 --es-timeout       time of inactivy before which a PID is reported down (in ms)" );
     msg_Raw( NULL, "  -r --remote-socket <remote socket>" );
-    msg_Raw( NULL, "  -V --version          only display the version" );
     msg_Raw( NULL, "  -Z --mrtg-file <file> Log input packets and errors into mrtg-file" );
+    msg_Raw( NULL, "  -V --version          only display the version" );
     exit(1);
 }
 
 int main( int i_argc, char **pp_argv )
 {
     const char *psz_network_name = "DVBlast - http://www.videolan.org/projects/dvblast.html";
-    char *p_network_name_tmp = NULL;
-    size_t i_network_name_tmp_size;
+    const char *psz_provider_name = NULL;
     char *psz_dup_config = NULL;
-    mtime_t i_poll_timeout = MAX_POLL_TIMEOUT;
     struct sched_param param;
     int i_error;
     int c;
-    struct sigaction sa;
-    sigset_t set;
-
     int b_enable_syslog = 0;
+    struct ev_signal sigint_watcher, sigterm_watcher, sighup_watcher;
+    struct ev_timer quit_watcher;
+
+    print_fh = stdout;
 
     if ( i_argc == 1 )
         usage();
 
     /*
-     * The only short options left are: ky0123456789
+     * The only short options left are: 489
      * Use them wisely.
      */
     static const struct option long_options[] =
@@ -542,17 +712,20 @@ int main( int i_argc, char **pp_argv )
         { "priority",        required_argument, NULL, 'i' },
         { "adapter",         required_argument, NULL, 'a' },
         { "frontend-number", required_argument, NULL, 'n' },
+        { "delsys",          required_argument, NULL, '5' },
         { "frequency",       required_argument, NULL, 'f' },
         { "fec-inner",       required_argument, NULL, 'F' },
         { "rolloff",         required_argument, NULL, 'R' },
         { "symbol-rate",     required_argument, NULL, 's' },
         { "diseqc",          required_argument, NULL, 'S' },
+        { "uncommitted",     required_argument, NULL, 'k' },
         { "voltage",         required_argument, NULL, 'v' },
         { "force-pulse",     no_argument,       NULL, 'p' },
         { "bandwidth",       required_argument, NULL, 'b' },
         { "inversion",       required_argument, NULL, 'I' },
         { "modulation",      required_argument, NULL, 'm' },
         { "pilot",           required_argument, NULL, 'P' },
+        { "multistream-id",  required_argument, NULL, '1' },
         { "fec-lp",          required_argument, NULL, 'K' },
         { "guard",           required_argument, NULL, 'G' },
         { "hierarchy",       required_argument, NULL, 'H' },
@@ -565,6 +738,7 @@ int main( int i_argc, char **pp_argv )
         { "latency",         required_argument, NULL, 'L' },
         { "retention",       required_argument, NULL, 'E' },
         { "duplicate",       required_argument, NULL, 'd' },
+        { "passthrough",     no_argument,       NULL, '3' },
         { "rtp-input",       required_argument, NULL, 'D' },
         { "asi-adapter",     required_argument, NULL, 'A' },
         { "any-type",        no_argument,       NULL, 'z' },
@@ -581,14 +755,19 @@ int main( int i_argc, char **pp_argv )
         { "logger-ident",    required_argument, NULL, 'g' },
         { "print",           required_argument, NULL, 'x' },
         { "quit-timeout",    required_argument, NULL, 'Q' },
+        { "print-period",    required_argument, NULL, '6' },
+        { "es-timeout",      required_argument, NULL, '7' },
         { "quiet",           no_argument,       NULL, 'q' },
         { "help",            no_argument,       NULL, 'h' },
         { "version",         no_argument,       NULL, 'V' },
         { "mrtg-file",       required_argument, NULL, 'Z' },
+        { "ca-number",       required_argument, NULL, 'y' },
+        { "pidmap",          required_argument, NULL, '0' },
+        { "dvr-buf-size",    required_argument, NULL, '2' },
         { 0, 0, 0, 0 }
     };
 
-    while ( (c = getopt_long(i_argc, pp_argv, "q::c:r:t:o:i:a:n:f:F:R:s:S:v:pb:I:m:P:K:G:H:X:O:uwUTL:E:d:D:A:lg:zCWYeM:N:j:J:B:x:Q:hVZ:", long_options, NULL)) != -1 )
+    while ( (c = getopt_long(i_argc, pp_argv, "q::c:r:t:o:i:a:n:5:f:F:R:s:S:k:v:pb:I:m:P:K:G:H:X:O:uwUTL:E:d:3D:A:lg:zCWYeM:N:j:J:B:x:Q:6:7:hVZ:y:0:1:2:", long_options, NULL)) != -1 )
     {
         switch ( c )
         {
@@ -617,6 +796,12 @@ int main( int i_argc, char **pp_argv )
 
         case 'c':
             psz_conf_file = optarg;
+            /*
+             * When configuration file is used it is reasonable to assume that
+             * services may be added/removed. If b_select_pmts is not set dvblast
+             * is unable to start streaming newly added services in the config.
+             */
+            b_select_pmts = 1;
             break;
 
         case 'r':
@@ -648,15 +833,28 @@ int main( int i_argc, char **pp_argv )
             i_fenum = strtol( optarg, NULL, 0 );
             break;
 
+        case 'y':
+            i_canum = strtol( optarg, NULL, 0 );
+            break;
+
+        case '5':
+            psz_delsys = optarg;
+            break;
+
         case 'f':
-            i_frequency = strtol( optarg, NULL, 0 );
+            if (optarg && optarg[0] != '-')
+                i_frequency = strtol( optarg, NULL, 0 );
             if ( pf_Open != NULL )
                 usage();
+#ifdef HAVE_DVB_SUPPORT
             pf_Open = dvb_Open;
-            pf_Read = dvb_Read;
             pf_Reset = dvb_Reset;
             pf_SetFilter = dvb_SetFilter;
             pf_UnsetFilter = dvb_UnsetFilter;
+#else
+            msg_Err( NULL, "DVBlast is compiled without DVB support.");
+            exit(1);
+#endif
             break;
 
         case 'F':
@@ -675,6 +873,10 @@ int main( int i_argc, char **pp_argv )
             i_satnum = strtol( optarg, NULL, 16 );
             break;
 
+        case 'k':
+            i_uncommitted = strtol( optarg, NULL, 16 );
+            break;
+
         case 'v':
             i_voltage = strtol( optarg, NULL, 0 );
             break;
@@ -699,6 +901,10 @@ int main( int i_argc, char **pp_argv )
             i_pilot = strtol( optarg, NULL, 0 );
             break;
 
+        case '1':
+            i_mis = strtol( optarg, NULL, 0 );
+            break;
+
         case 'K':
             i_fec_lp = strtol( optarg, NULL, 0 );
             break;
@@ -724,7 +930,7 @@ int main( int i_argc, char **pp_argv )
             break;
 
         case 'w':
-            b_select_pmts = 1;
+            b_select_pmts = !b_select_pmts;
             break;
 
         case 'U':
@@ -743,26 +949,50 @@ int main( int i_argc, char **pp_argv )
             psz_dup_config = optarg;
             break;
 
+        case '3':
+            b_passthrough = true;
+            print_fh = stderr;
+            break;
+
         case 'D':
             psz_udp_src = optarg;
             if ( pf_Open != NULL )
                 usage();
             pf_Open = udp_Open;
-            pf_Read = udp_Read;
             pf_Reset = udp_Reset;
             pf_SetFilter = udp_SetFilter;
             pf_UnsetFilter = udp_UnsetFilter;
             break;
 
         case 'A':
-            i_asi_adapter = strtol( optarg, NULL, 0 );
+#ifdef HAVE_ASI_SUPPORT
             if ( pf_Open != NULL )
                 usage();
-            pf_Open = asi_Open;
-            pf_Read = asi_Read;
-            pf_Reset = asi_Reset;
-            pf_SetFilter = asi_SetFilter;
-            pf_UnsetFilter = asi_UnsetFilter;
+            if ( strncmp(optarg, "deltacast:", 10) == 0)
+            {
+#ifdef HAVE_ASI_DELTACAST_SUPPORT
+                i_asi_adapter = strtol( optarg+10, NULL, 0 );
+                pf_Open = asi_deltacast_Open;
+                pf_Reset = asi_deltacast_Reset;
+                pf_SetFilter = asi_deltacast_SetFilter;
+                pf_UnsetFilter = asi_deltacast_UnsetFilter;
+#else
+                msg_Err( NULL, "DVBlast is compiled without Deltacast ASI support.");
+                exit(1);
+#endif
+            }
+            else
+            {
+                i_asi_adapter = strtol( optarg, NULL, 0 );
+                pf_Open = asi_Open;
+                pf_Reset = asi_Reset;
+                pf_SetFilter = asi_SetFilter;
+                pf_UnsetFilter = asi_UnsetFilter;
+            }
+#else
+            msg_Err( NULL, "DVBlast is compiled without ASI support.");
+            exit(1);
+#endif
             break;
 
         case 'z':
@@ -779,7 +1009,6 @@ int main( int i_argc, char **pp_argv )
 
         case 'Y':
             b_enable_ecm = true;
-            b_dvb_global = true;
             break;
  
         case 'e':
@@ -819,20 +1048,30 @@ int main( int i_argc, char **pp_argv )
             break;
 
         case 'x':
+            b_print_enabled = true;
             if ( !strcmp(optarg, "text") )
                 i_print_type = PRINT_TEXT;
             else if ( !strcmp(optarg, "xml") )
                 i_print_type = PRINT_XML;
             else
+            {
+                b_print_enabled = false;
                 msg_Warn( NULL, "unrecognized print type %s", optarg );
-            /* Make stdout line-buffered */
-            setvbuf(stdout, NULL, _IOLBF, 0);
+            }
             break;
 
         case 'Q':
             i_quit_timeout_duration = strtoll( optarg, NULL, 0 ) * 1000;
             break;
 
+        case '6':
+            i_print_period = strtoll( optarg, NULL, 0 ) * 1000;
+            break;
+
+        case '7':
+            i_es_timeout = strtoll( optarg, NULL, 0 ) * 1000;
+            break;
+
         case 'V':
             DisplayVersion();
             exit(0);
@@ -842,6 +1081,39 @@ int main( int i_argc, char **pp_argv )
             psz_mrtg_file = optarg;
             break;
 
+        case '0': {
+            /* We expect a comma separated list of numbers.
+               Put them into the pi_newpids array as they appear */
+            char *str1;
+            char *saveptr = NULL;
+            char *tok = NULL;
+            int i, i_newpid;
+            for (i = 0, str1 = optarg; i < N_MAP_PIDS; i++, str1 = NULL)
+            {
+                tok = strtok_r(str1, ",", &saveptr);
+                if ( !tok )
+                    break;
+                i_newpid = strtoul(tok, NULL, 0);
+                if ( !i_newpid ) {
+                     msg_Err( NULL, "Invalid pidmap string" );
+                     usage();
+                }
+                pi_newpids[i] = i_newpid;
+            }
+            b_do_remap = true;
+            break;
+        }
+#ifdef HAVE_DVB_SUPPORT
+        case '2':
+            i_dvr_buffer_size = strtol( optarg, NULL, 0 );
+            if (!i_dvr_buffer_size)
+                usage();	// it exits
+            /* roundup to packet size */
+            i_dvr_buffer_size += TS_SIZE - 1;
+            i_dvr_buffer_size /= TS_SIZE;
+            i_dvr_buffer_size *= TS_SIZE;
+            break;
+#endif
         case 'h':
         default:
             usage();
@@ -853,15 +1125,22 @@ int main( int i_argc, char **pp_argv )
     if ( b_enable_syslog )
         msg_Connect( psz_syslog_ident ? psz_syslog_ident : pp_argv[0] );
 
+    if ( b_print_enabled )
+    {
+        /* Make std* line-buffered */
+        setvbuf(print_fh, NULL, _IOLBF, 0);
+    }
+
     if ( i_verbose )
         DisplayVersion();
 
     msg_Warn( NULL, "restarting" );
+
     switch (i_print_type) 
     {
         case PRINT_XML:
-            printf("<?xml version=\"1.0\" encoding=\"utf-8\"?>\n");
-            printf("<TS>\n");
+            fprintf(print_fh, "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n");
+            fprintf(print_fh, "<TS>\n");
             break;
         default:
             break;
@@ -879,6 +1158,12 @@ int main( int i_argc, char **pp_argv )
         b_dvb_global = true;
     }
 
+    if ((event_loop = ev_default_loop(0)) == NULL)
+    {
+        msg_Err( NULL, "unable to initialize libev" );
+        exit(EXIT_FAILURE);
+    }
+
     memset( &output_dup, 0, sizeof(output_dup) );
     if ( psz_dup_config != NULL )
     {
@@ -896,53 +1181,13 @@ int main( int i_argc, char **pp_argv )
         config_Free( &config );
     }
 
-    if ( strcasecmp( psz_native_charset, psz_dvb_charset ) )
-    {
-#ifdef HAVE_ICONV
-        iconv_t iconv_system = iconv_open( psz_dvb_charset,
-                                           psz_native_charset );
-        if ( iconv_system != (iconv_t)-1 )
-        {
-            size_t i = strlen( psz_network_name );
-            char *p, *psz_string;
-            i_network_name_tmp_size = i * 6;
-            p = psz_string = malloc(i_network_name_tmp_size);
-            if ( iconv( iconv_system, (char **)&psz_network_name, &i, &p,
-                        &i_network_name_tmp_size ) == -1 )
-                free( psz_string );
-            else
-            {
-                p_network_name_tmp = psz_string;
-                i_network_name_tmp_size = p - psz_string;
-            }
-            iconv_close( iconv_system );
-        }
-#else
-        msg_Warn( NULL,
-                  "unable to convert from %s to %s (iconv is not available)",
-                  psz_native_charset, psz_dvb_charset );
-#endif
-    }
-    if ( p_network_name_tmp == NULL )
-    {
-        p_network_name_tmp = strdup(psz_network_name);
-        i_network_name_tmp_size = strlen(psz_network_name);
-    }
-    p_network_name = dvb_string_set( (uint8_t *)p_network_name_tmp,
-                                     i_network_name_tmp_size, psz_dvb_charset,
-                                     &i_network_name_size );
-    free( p_network_name_tmp );
+    config_strdvb( &network_name, psz_network_name );
+    config_strdvb( &provider_name, psz_provider_name );
 
     /* Set signal handlers */
-    memset( &sa, 0, sizeof(struct sigaction) );
-    sa.sa_handler = SigHandler;
-    sigfillset( &set );
-
-    if ( sigaction( SIGHUP, &sa, NULL ) == -1 || sigaction( SIGINT, &sa, NULL ) == -1 )
-    {
-        msg_Err( NULL, "couldn't set signal handler: %s", strerror(errno) );
-        exit(EXIT_FAILURE);
-    }
+    signal_watcher_init(&sigint_watcher, event_loop, sighandler, SIGINT);
+    signal_watcher_init(&sigterm_watcher, event_loop, sighandler, SIGTERM);
+    signal_watcher_init(&sighup_watcher, event_loop, sighandler, SIGHUP);
 
     srand( time(NULL) * getpid() );
 
@@ -963,62 +1208,44 @@ int main( int i_argc, char **pp_argv )
         }
     }
 
-    config_ReadFile( psz_conf_file );
+    config_ReadFile();
 
     if ( psz_srv_socket != NULL )
         comm_Open();
 
-    for ( ; ; )
+    if ( i_quit_timeout_duration )
     {
-        block_t *p_ts;
-
-        if ( b_exit_now )
-        {
-            msg_Info( NULL, "Shutdown was requested." );
-            break;
-        }
-
-        if ( b_conf_reload )
-        {
-            b_conf_reload = 0;
-            msg_Info( NULL, "Configuration reload was requested." );
-            config_ReadFile( psz_conf_file );
-        }
-
-        if ( i_quit_timeout && i_quit_timeout <= i_wallclock )
-        {
-            switch (i_print_type)
-            {
-            case PRINT_XML:
-                printf("</TS>\n");
-                break;
-            default:
-                break;
-            }
-            exit(EXIT_SUCCESS);
-        }
-
-        p_ts = pf_Read( i_poll_timeout );
-        if ( p_ts != NULL )
-        {
-            mrtgAnalyse(p_ts);
-            demux_Run( p_ts );
-        }
-        i_poll_timeout = output_Send();
-        if ( i_poll_timeout == -1 || i_poll_timeout > MAX_POLL_TIMEOUT )
-            i_poll_timeout = MAX_POLL_TIMEOUT;
+        ev_timer_init(&quit_watcher, quit_cb,
+                      i_quit_timeout_duration / 1000000., 0);
+        ev_timer_start(event_loop, &quit_watcher);
     }
 
+    outputs_Init();
+
+    ev_run(event_loop, 0);
+
     mrtgClose();
     outputs_Close( i_nb_outputs );
     demux_Close();
-    free( p_network_name );
+    dvb_string_clean( &network_name );
+    dvb_string_clean( &provider_name );
+    if ( conf_iconv != (iconv_t)-1 )
+        iconv_close( conf_iconv );
+
+    switch (i_print_type)
+    {
+    case PRINT_XML:
+        fprintf(print_fh, "</TS>\n");
+        break;
+    default:
+        break;
+    }
 
     if ( b_enable_syslog )
         msg_Disconnect();
 
-    if ( psz_srv_socket && i_comm_fd > -1 )
-        unlink( psz_srv_socket );
+    comm_Close();
+    block_Vacuum();
 
     return EXIT_SUCCESS;
 }
