@@ -1,7 +1,7 @@
 /*****************************************************************************
  * dvblast.h
  *****************************************************************************
- * Copyright (C) 2004, 2008-2011 VideoLAN
+ * Copyright (C) 2004, 2008-2011, 2015 VideoLAN
  *
  * Authors: Christophe Massiot <massiot@via.ecp.fr>
  *          Andy Gatward <a.j.gatward@reading.ac.uk>
@@ -23,25 +23,59 @@
 
 #include <netdb.h>
 #include <sys/socket.h>
+#include <netinet/udp.h>
+#include <netinet/ip.h>
 
-#define HAVE_CLOCK_NANOSLEEP
-#define HAVE_ICONV
+#include "config.h"
 
-#define DEFAULT_PORT 3001
-#define TS_SIZE 188
-#define MAX_PIDS 8192
-#define DEFAULT_IPV4_MTU 1500
-#define DEFAULT_IPV6_MTU 1280
-#define PADDING_PID 8191
-#define WATCHDOG_WAIT 10000000LL
-#define MAX_ERRORS 1000
-#define DEFAULT_VERBOSITY 4
-#define MAX_POLL_TIMEOUT 100000 /* 100 ms */
-#define DEFAULT_OUTPUT_LATENCY 200000 /* 200 ms */
-#define DEFAULT_MAX_RETENTION 40000 /* 40 ms */
-#define MAX_EIT_RETENTION 500000 /* 500 ms */
-#define DEFAULT_FRONTEND_TIMEOUT 30000000 /* 30 s */
-#define EXIT_STATUS_FRONTEND_TIMEOUT 100
+#ifndef container_of
+#   define container_of(ptr, type, member) ({                               \
+        const typeof( ((type *)0)->member ) *_mptr = (ptr);                 \
+        (type *)( (char *)_mptr - offsetof(type,member) );})
+#endif
+
+/* Defines for pid mapping */
+#define N_MAP_PIDS                 4
+/* Offsets in the command line args for the pid mapping */
+typedef enum
+{
+    I_PMTPID = 0, I_APID, I_VPID, I_SPUPID
+} pidmap_offset;
+
+/* Impossible PID value */
+#define UNUSED_PID (MAX_PIDS + 1)
+
+/*****************************************************************************
+ * Raw udp packet structure with flexible-array payload
+ *****************************************************************************/
+struct udpheader { // FAVOR_BSD hell ...
+  u_int16_t source;
+  u_int16_t dest;
+  u_int16_t len;
+  u_int16_t check;
+};
+
+#if defined(__FreeBSD__) || defined(__APPLE__)
+struct iphdr {
+    unsigned int ihl:4;
+    unsigned int version:4;
+    uint8_t tos;
+    uint16_t tot_len;
+    uint16_t id;
+    uint16_t frag_off;
+    uint8_t ttl;
+    uint8_t protocol;
+    uint16_t check;
+    uint32_t saddr;
+    uint32_t daddr;
+};
+#endif
+
+struct udprawpkt {
+    struct  iphdr iph;
+    struct  udpheader udph;
+    uint8_t payload[];
+} __attribute__((packed));
 
 /*****************************************************************************
  * Output configuration flags (for output_t -> i_config) - bit values
@@ -52,6 +86,7 @@
  * Bit  4 : Set for file / FIFO output, unset for network (future use)
  * Bit  5 : Set if DVB conformance tables are inserted
  * Bit  6 : Set if DVB EIT schedule tables are forwarded
+ * Bit  7 : Set for RAW socket output
  *****************************************************************************/
 
 #define OUTPUT_WATCH         0x01
@@ -61,6 +96,7 @@
 #define OUTPUT_FILE          0x10
 #define OUTPUT_DVB           0x20
 #define OUTPUT_EPG           0x40
+#define OUTPUT_RAW           0x80
 
 typedef int64_t mtime_t;
 
@@ -69,11 +105,18 @@ typedef struct block_t
     uint8_t p_ts[TS_SIZE];
     int i_refcount;
     mtime_t i_dts;
+    uint16_t tmp_pid;
     struct block_t *p_next;
 } block_t;
 
 typedef struct packet_t packet_t;
 
+typedef struct dvb_string_t
+{
+    uint8_t *p;
+    size_t i;
+} dvb_string_t;
+
 typedef struct output_config_t
 {
     /* identity */
@@ -87,19 +130,29 @@ typedef struct output_config_t
     uint64_t i_config;
 
     /* output config */
-    char *psz_service_name;
-    char *psz_service_provider;
+    uint16_t i_network_id;
+    dvb_string_t network_name;
+    dvb_string_t service_name;
+    dvb_string_t provider_name;
     uint8_t pi_ssrc[4];
     mtime_t i_output_latency, i_max_retention;
     int i_ttl;
     uint8_t i_tos;
     int i_mtu;
+    char *psz_srcaddr; /* raw packets */
+    int i_srcport;
 
     /* demux config */
     int i_tsid;
     uint16_t i_sid; /* 0 if raw mode */
     uint16_t *pi_pids;
     int i_nb_pids;
+    uint16_t i_new_sid;
+    bool b_passthrough;
+
+    /* for pidmap from config file */
+    bool b_do_remap;
+    uint16_t pi_confpids[N_MAP_PIDS];
 } output_config_t;
 
 typedef struct output_t
@@ -109,9 +162,9 @@ typedef struct output_t
     /* output */
     int i_handle;
     packet_t *p_packets, *p_last_packet;
+    packet_t *p_packet_lifo;
+    unsigned int i_packet_count;
     uint16_t i_seqnum;
-    mtime_t i_ref_timestamp;
-    mtime_t i_ref_wallclock;
 
     /* demux */
     int i_nb_errors;
@@ -127,6 +180,12 @@ typedef struct output_t
     block_t *p_eit_ts_buffer;
     uint8_t i_eit_ts_buffer_offset, i_eit_cc;
     uint16_t i_tsid;
+    // Arrays used for mapping pids.
+    // newpids is indexed using the original pid
+    uint16_t pi_newpids[MAX_PIDS];
+    uint16_t pi_freepids[MAX_PIDS];   // used where multiple streams of the same type are used
+
+    struct udprawpkt raw_pkt_header;
 } output_t;
 
 typedef struct ts_pid_info {
@@ -143,18 +202,23 @@ typedef struct ts_pid_info {
        3 = Scrambled with odd key */
 } ts_pid_info_t;
 
+extern struct ev_loop *event_loop;
 extern int i_syslog;
 extern int i_verbose;
 extern output_t **pp_outputs;
 extern int i_nb_outputs;
 extern output_t output_dup;
+extern bool b_passthrough;
 extern char *psz_srv_socket;
-extern int i_comm_fd;
 extern int i_adapter;
 extern int i_fenum;
+extern int i_canum;
+extern char *psz_delsys;
+extern int i_dvr_buffer_size;
 extern int i_frequency;
 extern int i_srate;
 extern int i_satnum;
+extern int i_uncommitted;
 extern int i_fec;
 extern int i_rolloff;
 extern int i_voltage;
@@ -163,32 +227,35 @@ extern int i_bandwidth;
 extern int i_inversion;
 extern char *psz_modulation;
 extern int i_pilot;
+extern int i_mis;
 extern int i_fec_lp;
 extern int i_guard;
 extern int i_transmission;
 extern int i_hierarchy;
 extern mtime_t i_frontend_timeout_duration;
-extern mtime_t i_quit_timeout;
 extern mtime_t i_quit_timeout_duration;
 extern int b_budget_mode;
 extern int b_any_type;
 extern int b_select_pmts;
 extern int b_random_tsid;
-extern uint16_t i_network_id;
-extern uint8_t *p_network_name;
-extern size_t i_network_name_size;
+extern bool b_enable_emm;
+extern bool b_enable_ecm;
 extern mtime_t i_wallclock;
-extern volatile int b_conf_reload;
-extern volatile int b_exit_now;
-extern int i_comm_fd;
 extern char *psz_udp_src;
 extern int i_asi_adapter;
 extern const char *psz_native_charset;
-extern const char *psz_dvb_charset;
 extern enum print_type_t i_print_type;
+extern bool b_print_enabled;
+extern FILE *print_fh;
+extern mtime_t i_print_period;
+extern mtime_t i_es_timeout;
+
+/* pid mapping */
+extern bool b_do_remap;
+extern uint16_t pi_newpids[N_MAP_PIDS];
+extern void init_pid_mapping( output_t * );
 
 extern void (*pf_Open)( void );
-extern block_t * (*pf_Read)( mtime_t i_poll_timeout );
 extern void (*pf_Reset)( void );
 extern int (*pf_SetFilter)( uint16_t i_pid );
 extern void (*pf_UnsetFilter)( int i_fd, uint16_t i_pid );
@@ -199,7 +266,6 @@ extern void (*pf_UnsetFilter)( int i_fd, uint16_t i_pid );
 
 void config_Init( output_config_t *p_config );
 void config_Free( output_config_t *p_config );
-bool config_ParseHost( output_config_t *p_config, char *psz_string );
 
 /* Connect/Disconnect from syslogd */
 void msg_Connect( const char *ident );
@@ -215,11 +281,19 @@ __attribute__ ((format(printf, 2, 3))) void msg_Raw( void *_unused, const char *
 /* */
 bool streq(char *a, char *b);
 char * xstrdup(char *str);
+
+void dvb_string_init(dvb_string_t *p_dvb_string);
+void dvb_string_clean(dvb_string_t *p_dvb_string);
+void dvb_string_copy(dvb_string_t *p_dst, const dvb_string_t *p_src);
+int dvb_string_cmp(const dvb_string_t *p_1, const dvb_string_t *p_2);
+
 mtime_t mdate( void );
 void msleep( mtime_t delay );
 void hexDump( uint8_t *p_data, uint32_t i_len );
 struct addrinfo *ParseNodeService( char *_psz_string, char **ppsz_end,
                                    uint16_t i_default_port );
+char *config_stropt( const char *psz_string );
+void config_ReadFile(void);
 
 uint8_t *psi_pack_section( uint8_t *p_sections, unsigned int *pi_size );
 uint8_t *psi_pack_sections( uint8_t **pp_sections, unsigned int *pi_size );
@@ -227,23 +301,27 @@ uint8_t **psi_unpack_sections( uint8_t *p_flat_sections, unsigned int i_size );
 
 void dvb_Open( void );
 void dvb_Reset( void );
-block_t * dvb_Read( mtime_t i_poll_timeout );
 int dvb_SetFilter( uint16_t i_pid );
 void dvb_UnsetFilter( int i_fd, uint16_t i_pid );
 uint8_t dvb_FrontendStatus( uint8_t *p_answer, ssize_t *pi_size );
 
 void udp_Open( void );
-block_t * udp_Read( mtime_t i_poll_timeout );
 void udp_Reset( void );
 int udp_SetFilter( uint16_t i_pid );
 void udp_UnsetFilter( int i_fd, uint16_t i_pid );
 
 void asi_Open( void );
-block_t * asi_Read( mtime_t i_poll_timeout );
 void asi_Reset( void );
 int asi_SetFilter( uint16_t i_pid );
 void asi_UnsetFilter( int i_fd, uint16_t i_pid );
 
+#ifdef HAVE_ASI_DELTACAST_SUPPORT
+void asi_deltacast_Open( void );
+void asi_deltacast_Reset( void );
+int asi_deltacast_SetFilter( uint16_t i_pid );
+void asi_deltacast_UnsetFilter( int i_fd, uint16_t i_pid );
+#endif
+
 void demux_Open( void );
 void demux_Run( block_t *p_ts );
 void demux_Change( output_t *p_output, const output_config_t *p_config );
@@ -265,32 +343,17 @@ output_t *output_Create( const output_config_t *p_config );
 int output_Init( output_t *p_output, const output_config_t *p_config );
 void output_Close( output_t *p_output );
 void output_Put( output_t *p_output, block_t *p_block );
-mtime_t output_Send( void );
 output_t *output_Find( const output_config_t *p_config );
 void output_Change( output_t *p_output, const output_config_t *p_config );
+void outputs_Init( void );
 void outputs_Close( int i_num_outputs );
 
 void comm_Open( void );
-void comm_Read( void );
+void comm_Close( void );
 
-/*****************************************************************************
- * block_New
- *****************************************************************************/
-static inline block_t *block_New( void )
-{
-    block_t *p_block = malloc(sizeof(block_t));
-    p_block->p_next = NULL;
-    p_block->i_refcount = 1;
-    return p_block;
-}
-
-/*****************************************************************************
- * block_Delete
- *****************************************************************************/
-static inline void block_Delete( block_t *p_block )
-{
-    free( p_block );
-}
+block_t *block_New( void );
+void block_Delete( block_t *p_block );
+void block_Vacuum( void );
 
 /*****************************************************************************
  * block_DeleteChain
@@ -300,7 +363,7 @@ static inline void block_DeleteChain( block_t *p_block )
     while ( p_block != NULL )
     {
         block_t *p_next = p_block->p_next;
-        free( p_block );
+        block_Delete( p_block );
         p_block = p_next;
     }
 }
