@@ -38,6 +38,8 @@
 #include <errno.h>
 #include <getopt.h>
 
+#include <iconv.h>
+
 #include <bitstream/mpeg/psi.h>
 #include <bitstream/dvb/si.h>
 #include <bitstream/dvb/si_print.h>
@@ -56,6 +58,8 @@ mtime_t now;
 int i_fd = -1;
 char psz_client_socket[PATH_MAX] = {0};
 
+static iconv_t iconv_handle = (iconv_t)-1;
+
 static void clean_client_socket() {
     if ( i_fd > -1 )
     {
@@ -82,6 +86,7 @@ static void psi_print(void *_unused, const char *psz_format, ...)
     strcpy(psz_fmt, psz_format);
     strcat(psz_fmt, "\n");
     vprintf(psz_fmt, args);
+    va_end(args);
 }
 
 __attribute__ ((format(printf, 1, 2)))
@@ -99,6 +104,7 @@ void return_error( const char *psz_format, ... )
         snprintf( psz_fmt, sizeof(psz_fmt) - 1, "ERROR: %s\n", psz_format );
     psz_fmt[sizeof(psz_fmt) - 1] = '\0';
     vfprintf( stderr, psz_fmt, args );
+    va_end(args);
     exit(255);
 }
 
@@ -110,10 +116,49 @@ static char *iconv_append_null(const char *p_string, size_t i_length)
     return psz_string;
 }
 
+const char *psz_native_charset = "UTF-8";
+
 char *psi_iconv(void *_unused, const char *psz_encoding,
                   char *p_string, size_t i_length)
 {
-    return iconv_append_null(p_string, i_length);
+    static const char *psz_current_encoding = "";
+
+    char *psz_string, *p;
+    size_t i_out_length;
+
+    if (!strcmp(psz_encoding, psz_native_charset))
+        return iconv_append_null(p_string, i_length);
+
+    if (iconv_handle != (iconv_t)-1 &&
+        strcmp(psz_encoding, psz_current_encoding)) {
+        iconv_close(iconv_handle);
+        iconv_handle = (iconv_t)-1;
+    }
+
+    if (iconv_handle == (iconv_t)-1)
+        iconv_handle = iconv_open(psz_native_charset, psz_encoding);
+    if (iconv_handle == (iconv_t)-1) {
+        msg_Warn(NULL, "couldn't open converter from %s to %s (%m)", psz_encoding,
+                psz_native_charset);
+        return iconv_append_null(p_string, i_length);
+    }
+    psz_current_encoding = psz_encoding;
+
+    /* converted strings can be up to six times larger */
+    i_out_length = i_length * 6;
+    p = psz_string = malloc(i_out_length);
+    if (iconv(iconv_handle, &p_string, &i_length, &p, &i_out_length) == -1) {
+        msg_Warn(NULL, "couldn't convert from %s to %s (%m)", psz_encoding,
+                psz_native_charset);
+        free(psz_string);
+        return iconv_append_null(p_string, i_length);
+    }
+    if (i_length)
+        msg_Warn(NULL, "partial conversion from %s to %s", psz_encoding,
+                psz_native_charset);
+
+    *p = '\0';
+    return psz_string;
 }
 
 void print_pids_header( void )
@@ -207,6 +252,7 @@ void usage()
     printf("Control commands:\n");
     printf("  reload                          Reload configuration.\n");
     printf("  shutdown                        Shutdown DVBlast.\n");
+#ifdef HAVE_DVB_SUPPORT
     printf("Status commands:\n");
     printf("  fe_status                       Read frontend status information.\n");
     printf("  mmi_status                      Read CAM status.\n");
@@ -217,6 +263,7 @@ void usage()
     printf("  mmi_get <slot>                  Read MMI slot.\n");
     printf("  mmi_send_text <slot> <text>     Send text to MMI slot.\n");
     printf("  mmi_send_choice <slot> <choice> Send choice to MMI slot.\n");
+#endif
     printf("Demux info commands:\n");
     printf("  get_pat                         Return last PAT table.\n");
     printf("  get_cat                         Return last CAT table.\n");
@@ -386,6 +433,7 @@ int main( int i_argc, char **ppsz_argv )
         p_data[1] = (uint8_t)(i_pid & 0xff);
         break;
     }
+#ifdef HAVE_DVB_SUPPORT
     case CMD_MMI_SEND_TEXT:
     {
         struct cmd_mmi_send *p_cmd = (struct cmd_mmi_send *)p_data;
@@ -432,6 +480,7 @@ int main( int i_argc, char **ppsz_argv )
         i_size = COMM_HEADER_SIZE + 1;
         break;
     }
+#endif
     default:
         /* This should not happen */
         return_error( "Unhandled option (%d)", opt.cmd );
@@ -498,6 +547,11 @@ int main( int i_argc, char **ppsz_argv )
         uint8_t *p_flat_data = p_buffer + COMM_HEADER_SIZE;
         unsigned int i_flat_data_size = i_size - COMM_HEADER_SIZE;
         uint8_t **pp_sections = psi_unpack_sections( p_flat_data, i_flat_data_size );
+        if ( !pp_sections )
+        {
+            return_error( "Error unpacking PSI" );
+            break;
+        }
 
         switch( c_answer )
         {
@@ -533,6 +587,7 @@ int main( int i_argc, char **ppsz_argv )
         break;
     }
 
+#ifdef HAVE_DVB_SUPPORT
     case RET_FRONTEND_STATUS:
     {
         int ret = 1;
@@ -568,6 +623,12 @@ int main( int i_argc, char **ppsz_argv )
         else \
             printf( "%s: %u\n", STRINGIFY(x), p_ret->info.x ); \
     } while(0)
+
+        if ( i_print_type == PRINT_XML )
+            printf( " <SETTING name=\"%s\"/>\n", p_ret->info.name );
+        else
+            printf( "name: %s\n", p_ret->info.name );
+
         PRINT_INFO( frequency_min );
         PRINT_INFO( frequency_max );
         PRINT_INFO( frequency_stepsize );
@@ -615,8 +676,6 @@ int main( int i_argc, char **ppsz_argv )
         PRINT_CAPS( CAN_HIERARCHY_AUTO );
         PRINT_CAPS( CAN_MUTE_TS );
 
-#define DVBAPI_VERSION ((DVB_API_VERSION)*100+(DVB_API_VERSION_MINOR))
-
 #if DVBAPI_VERSION >= 301
         PRINT_CAPS( CAN_8VSB );
         PRINT_CAPS( CAN_16VSB );
@@ -629,6 +688,9 @@ int main( int i_argc, char **ppsz_argv )
 #if DVBAPI_VERSION >= 501
         PRINT_CAPS( CAN_2G_MODULATION );
 #endif
+#if DVBAPI_VERSION >= 508
+        PRINT_CAPS( CAN_MULTISTREAM );
+#endif
 #undef PRINT_CAPS
 
         if ( i_print_type == PRINT_TEXT )
@@ -656,13 +718,13 @@ int main( int i_argc, char **ppsz_argv )
         {
             if ( i_print_type == PRINT_XML )
             {
-                printf(" <VALUE bit_error_rate=\"%d\"/>\n", p_ret->i_ber);
-                printf(" <VALUE signal_strength=\"%d\"/>\n", p_ret->i_strength);
-                printf(" <VALUE SNR=\"%d\"/>\n", p_ret->i_snr);
+                printf(" <VALUE bit_error_rate=\"%u\"/>\n", p_ret->i_ber);
+                printf(" <VALUE signal_strength=\"%u\"/>\n", p_ret->i_strength);
+                printf(" <VALUE SNR=\"%u\"/>\n", p_ret->i_snr);
             } else {
-                printf("\nBit error rate: %d\n", p_ret->i_ber);
-                printf("Signal strength: %d\n", p_ret->i_strength);
-                printf("SNR: %d\n", p_ret->i_snr);
+                printf("\nBit error rate: %u\n", p_ret->i_ber);
+                printf("Signal strength: %u\n", p_ret->i_strength);
+                printf("SNR: %u\n", p_ret->i_snr);
             }
             ret = 0;
         }
@@ -788,10 +850,16 @@ int main( int i_argc, char **ppsz_argv )
         exit(255);
         break;
     }
+#endif
 
     default:
         return_error( "Unknown command answer: %u", c_answer );
     }
 
+    if (iconv_handle != (iconv_t)-1) {
+        iconv_close(iconv_handle);
+        iconv_handle = (iconv_t)-1;
+    }
+
     return 0;
 }
